
/* This file is part of the Q programming system.

   The Q programming system is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   The Q programming system is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


#if defined (HAVE_CONFIG_H)
#  include "config.h"
#endif

/* system headers */

#ifdef USE_THREADS
/* make sure we also get the SUSV2 stuff (PTHREAD_MUTEX_RECURSIVE etc.) */
#ifndef _XOPEN_SOURCE
#define _XOPEN_SOURCE 500
#endif
#endif

#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <math.h>

#ifdef HAVE_ERRNO_H
#include <errno.h>
#else
# ifndef errno
extern int errno;
# endif
#endif

#include <sys/stat.h>

/* check for standard C headers */
#if STDC_HEADERS
# include <stdlib.h>
# include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <sys/types.h>
#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef USE_THREADS
#include <pthread.h>
#include <semaphore.h>
#endif

#ifdef WIN32
#include <windows.h>
#include <process.h>
#endif

/* make sure that gmp.h is included prior to libq.h */

#include <gmp.h>
#include <libq.h>

/* work around a bug in some gmp versions */

#define my_mpz_fits_slong_p(z) (mpz_size(z) == 0 || mpz_fits_slong_p(z))
#define my_mpz_fits_ulong_p(z) (mpz_size(z) == 0 || mpz_fits_ulong_p(z))
#define my_mpz_fits_uint_p(z) (mpz_size(z) == 0 || mpz_fits_uint_p(z))

#include "glob.h"
#include "regex.h"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

MODULE(clib)

#ifndef HAVE_STRDUP

char *strdup(char *s)
{
  char *t;
  return ((t=malloc(strlen(s)+1))?strcpy(t, s):NULL);
}

#endif

#ifndef HAVE_MEMCPY

#if __GNUC__ > 1
#define memcpy(TO,FROM,COUNT)	__builtin_memcpy(TO,FROM,COUNT)
#else
void *memcpy (to, from, count)
     char *to;
     char *from;
     int count;
{
  register char *f = from;
  register char *t = to;
  register int i = count;

  while (i-- > 0)
    *t++ = *f++;
  return (void*)to;
}
#endif

#endif

#define BUFSZ 1024

/* GMP convenience stuff */

#define BYTES_PER_LIMB sizeof(mp_limb_t)

static void*
my_mpz_realloc(m, new_size)
     mpz_ptr m;
     mp_size_t new_size;
{
  mpz_t m1;
  memcpy(m1, m, sizeof(mpz_t));
  if (_mpz_realloc(m, new_size))
    return m->_mp_d;
  else {
    if (m1->_mp_d) mpz_clear(m1);
    return NULL;
  }
}

static void *mpz_new(mpz_t z, mp_size_t sz)
{
  mpz_init(z);
  if (z->_mp_d && my_mpz_realloc(z, sz))
    return z->_mp_d;
  else
    return NULL;
}

static void *mpz_copy(mpz_t z, mpz_t u)
{
  if (mpz_new(z, mpz_size(u))) {
    mpz_set(z, u);
    return z->_mp_d;
  } else
    return NULL;
}

static void *mpz_resize(mpz_t z, mp_size_t sz)
{
  if (sz < mpz_size(z)) sz = mpz_size(z);
  if (sz == 0) sz = 1;
  if (sz != z->_mp_alloc && !my_mpz_realloc(z, sz))
    return NULL;
  else
    return z->_mp_d;
}

#define mpz_actsize(z) mpz_resize(z, mpz_size(z))

#define abs(x) (((x)<0.0)?-x:x)

static void *mpz_from_double(mpz_t z, double x)
{
  double ip, fp, dsz;
  int sz;
  fp = modf(x, &ip);
  /* estimate the number of limbs required */
  dsz = log(abs(ip))/log(2)/((double)CHAR_BIT*sizeof(mp_limb_t))+1.0;
  if (dsz < 1.0) dsz = 1.0; /* this can't happen?? */
  /* add an extra limb to be safe */
  sz = ((int)dsz)+1;
  /* this shouldn't happen but ... ;-) */
  if (((double)INT_MAX) <= dsz || sz < 0) return NULL;
  if (mpz_new(z, sz)) {
    mpz_set_d(z, x);
    return mpz_actsize(z);
  } else
    return NULL;
}

static long long_max(long x, long y)
{
  return (x > y)?x:y;
}

static long long_min(long x, long y)
{
  return (x < y)?x:y;
}

static void *mpz_addop2(void (*f)(), mpz_t z, mpz_t u, mpz_t v)
{
  int sz = long_max(mpz_size(u),mpz_size(v))+1;
  if (sz < 0) return NULL;
  if (mpz_new(z, sz)) {
    f(z, u, v);
    return mpz_actsize(z);
  } else
    return NULL;
}

static void *mpz_addop1(void (*f)(), mpz_t z, mpz_t u, mp_limb_t v)
{
  int sz = mpz_size(u)+1;
  if (sz < 0) return NULL;
  if (mpz_new(z, sz)) {
    f(z, u, v);
    return mpz_actsize(z);
  } else
    return NULL;
}

/* reverse the limbs of an mpz_t value to convert between little- and
   big-endian limb order */

static void reverse_limbs(mp_limb_t *p, unsigned n)
{
  unsigned i;
  for (i = 0; i < n/2; i++) {
    mp_limb_t x = p[i];
    p[i] = p[n-i-1];
    p[n-i-1] = x;
  }
}

/* libq doesn't provide these, so we do them ourselves */

static expr mkuint(unsigned long u)
{
  mpz_t z;
  mpz_init(z);
  if (z->_mp_d) {
    mpz_set_ui(z, u);
    return mkmpz(z);
  } else
    return NULL;
}

static int isuint(expr x, unsigned long *u)
{
  mpz_t z;
  if (ismpz(x, z) && mpz_sgn(z) >= 0 && my_mpz_fits_ulong_p(z)) {
    *u = mpz_get_ui(z);
    return 1;
  } else
    return 0;
}

#ifdef USE_THREADS
static pthread_mutex_t format_mutex;
#endif

static void lock_format(void)
{
#ifdef USE_THREADS
  pthread_mutex_lock(&format_mutex);
#endif
}

static void unlock_format(void)
{
#ifdef USE_THREADS
  pthread_mutex_unlock(&format_mutex);
#endif
}

/* additional string functions: *******************************************/

FUNCTION(clib,islower,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s) && s[0] && !s[1])
    if (islower(*s))
      return mktrue;
    else
      return mkfalse;
  else
    return __FAIL;
}

FUNCTION(clib,isupper,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s) && s[0] && !s[1])
    if (isupper(*s))
      return mktrue;
    else
      return mkfalse;
  else
    return __FAIL;
}

FUNCTION(clib,isalpha,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s) && s[0] && !s[1])
    if (isalpha(*s))
      return mktrue;
    else
      return mkfalse;
  else
    return __FAIL;
}

FUNCTION(clib,isdigit,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s) && s[0] && !s[1])
    if (isdigit(*s))
      return mktrue;
    else
      return mkfalse;
  else
    return __FAIL;
}

FUNCTION(clib,isxdigit,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s) && s[0] && !s[1])
    if (isxdigit(*s))
      return mktrue;
    else
      return mkfalse;
  else
    return __FAIL;
}

FUNCTION(clib,isalnum,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s) && s[0] && !s[1])
    if (isalnum(*s))
      return mktrue;
    else
      return mkfalse;
  else
    return __FAIL;
}

FUNCTION(clib,ispunct,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s) && s[0] && !s[1])
    if (ispunct(*s))
      return mktrue;
    else
      return mkfalse;
  else
    return __FAIL;
}

FUNCTION(clib,isspace,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s) && s[0] && !s[1])
    if (isspace(*s))
      return mktrue;
    else
      return mkfalse;
  else
    return __FAIL;
}

FUNCTION(clib,isgraph,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s) && s[0] && !s[1])
    if (isgraph(*s))
      return mktrue;
    else
      return mkfalse;
  else
    return __FAIL;
}

FUNCTION(clib,isprint,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s) && s[0] && !s[1])
    if (isprint(*s))
      return mktrue;
    else
      return mkfalse;
  else
    return __FAIL;
}

FUNCTION(clib,iscntrl,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s) && s[0] && !s[1])
    if (iscntrl(*s))
      return mktrue;
    else
      return mkfalse;
  else
    return __FAIL;
}

FUNCTION(clib,isascii,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s) && s[0] && !s[1])
    if (isascii(*s))
      return mktrue;
    else
      return mkfalse;
  else
    return __FAIL;
}

FUNCTION(clib,tolower,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s)) {
    char *t = strdup(s);
    if (t) {
      int i;
      for (i = 0; i < strlen(t); i++) t[i] = tolower(t[i]);
      return mkstr(t);
    } else
      return __ERROR;
  } else
    return __FAIL;
}

FUNCTION(clib,toupper,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s)) {
    char *t = strdup(s);
    if (t) {
      int i;
      for (i = 0; i < strlen(t); i++) t[i] = toupper(t[i]);
      return mkstr(t);
    } else
      return __ERROR;
  } else
    return __FAIL;
}

/* enhanced file functions: ***********************************************/

FUNCTION(clib,fopen,argc,argv)
{
  char *name, *mode;
  FILE *fp;
  if (argc != 2 || !isstr(argv[0], &name) || !isstr(argv[1], &mode))
    return __FAIL;
  if (!strchr("rwa", mode[0]) ||
      mode[1] && !strchr("b+", mode[1]) ||
      mode[1] && mode[2] && (mode[1] == mode[2] || !strchr("b+", mode[2])))
    return __FAIL;
  fp = fopen(name, mode);
  if (fp)
    return mkfile(fp);
  else
    return __FAIL;
}

FUNCTION(clib,freopen,argc,argv)
{
  char *name, *mode;
  FILE *fp;
  if (argc != 3 || !isstr(argv[0], &name) || !isstr(argv[1], &mode) ||
      !isfile(argv[2], &fp))
    return __FAIL;
  if (!strchr("rwa", mode[0]) ||
      mode[1] && !strchr("b+", mode[1]) ||
      mode[1] && mode[2] && (mode[1] == mode[2] || !strchr("b+", mode[2])))
    return __FAIL;
  if (freopen(name, mode, fp))
    return argv[2];
  else
    return __FAIL;
}

FUNCTION(clib,ftell,argc,argv)
{
  FILE *fp;
  long res;
  if (argc != 1 || !isfile(argv[0], &fp))
    return __FAIL;
  res = ftell(fp);
  if (res >= 0)
    return mkint(res);
  else
    return __FAIL;
}

FUNCTION(clib,fseek,argc,argv)
{
  FILE *fp;
  long pos;
  if (argc != 2 || !isfile(argv[0], &fp) || !isint(argv[1], &pos))
    return __FAIL;
  if (fseek(fp, pos, SEEK_SET))
    return __FAIL;
  else
    return mkvoid;
}

#ifdef __MINGW32__

/* win32 tmpnam is broken, provide a reasonable replacement */

#define tmpnam mytmpnam

static char *tmpnam(char *s)
{
  static char *t = NULL;
  char *p;
  if (t) free(t);
  t = _tempnam(NULL, "t");
  if (!t) return NULL;
  /* make sure to convert all \'s to /'s */
  while ((p = strchr(t, '\\'))) *p = '/';
  if (s) {
    strcpy(s, t);
    return s;
  } else
    return t;
}

#endif

FUNCTION(clib,tmpnam,argc,argv)
{
  if (argc == 0) {
    char s[L_tmpnam];
    tmpnam(s);
    return mkstr(strdup(s));
  } else
    return __FAIL;
}

FUNCTION(clib,tmpfile,argc,argv)
{
  if (argc == 0) {
    FILE * fp = tmpfile();
    if (fp)
      return mkfile(fp);
    else
      return __FAIL;
  } else
    return __FAIL;
}

/* binary file I/O: *******************************************************/

FUNCTION(clib,intsize,argc,argv)
{
  mpz_t z;
  if (argc == 1 && ismpz(argv[0], z)) {
    unsigned long n = mpz_size(z), i;
    mpz_t u;
    mpz_init(u);
    if (!u->_mp_d)
      return __ERROR;
    if (n != 0) {
      mp_limb_t x;
      for (x = z->_mp_d[n-1], i = 0; x != 0 && i < BYTES_PER_LIMB; i++)
	x >>= 8;
      if (z->_mp_d[n-1] & (1 << ((i<<3)-1))) i++;
      n = (n-1)*BYTES_PER_LIMB+i;
      mpz_set_ui(u, n);
    }
    return mkmpz(u);
  } else
    return __FAIL;
}

FUNCTION(clib,uintsize,argc,argv)
{
  mpz_t z;
  if (argc == 1 && ismpz(argv[0], z)) {
    unsigned long n = mpz_size(z), i;
    mpz_t u;
    mpz_init(u);
    if (!u->_mp_d)
      return __ERROR;
    if (n != 0) {
      mp_limb_t x;
      for (x = z->_mp_d[n-1], i = 0; x != 0 && i < BYTES_PER_LIMB; i++)
	x >>= 8;
      n = (n-1)*BYTES_PER_LIMB+i;
      mpz_set_ui(u, n);
    }
    return mkmpz(u);
  } else
    return __FAIL;
}

FUNCTION(clib,readint,argc,argv)
{
  FILE *fp;
  mpz_t c;
  if (argc == 2 && isfile(argv[0], &fp) && ismpz(argv[1], c) &&
      mpz_sgn(c) >= 0 && my_mpz_fits_uint_p(c)) {
    mpz_t u;
    unsigned long count = mpz_get_ui(c), n = count/BYTES_PER_LIMB,
      k = count-n*BYTES_PER_LIMB, l = 0, i;
    int sgn = 1;
    if (k != 0) {
      n++;
#ifdef WORDS_BIGENDIAN
      /* right-adjust on big-endian system */
      l = BYTES_PER_LIMB-k;
#endif
    }
    if (!mpz_new(u, n)) return __ERROR;
    for (i = 0; i < n; i++) u->_mp_d[i] = 0;
    /* read stuff */
    release_lock();
    if (read(fileno(fp), ((char*)u->_mp_d)+l, count) < count) {
      acquire_lock();
      mpz_clear(u);
      return __FAIL;
    }
    acquire_lock();
#ifdef WORDS_BIGENDIAN
    /* reverse limbs (GMP always orders limbs least significant first) */
    reverse_limbs(u->_mp_d, n);
#endif
    /* check for highest bit set (negative number) */
    if (k == 0) k = BYTES_PER_LIMB;
    if (n > 0 && (u->_mp_d[n-1] & (1 << ((k<<3)-1)))) {
      /* invert 2's complement */
      mp_limb_t x;
      sgn = -1;
      /* sub 1 */
      mpn_sub_1(u->_mp_d, u->_mp_d, n, 1U);
      /* 1's complement */
      for (i = 0; i < n-1; i++) u->_mp_d[i] = ~u->_mp_d[i];
      /* last limb with rest bytes treated separately */
      x = 0U;
      for (i = 0; i < k; i++) x = (x << 8) | 0xff;
      u->_mp_d[n-1] = ~u->_mp_d[n-1] & x;
    }
    /* adjust size */
    for (i = n; i > 0; i--)
      if (u->_mp_d[i-1] != 0) break;
    u->_mp_size = i*sgn;
    if (!mpz_actsize(u))
      return __FAIL;
    else
      return mkmpz(u);
  } else
    return __FAIL;
}

FUNCTION(clib,readuint,argc,argv)
{
  FILE *fp;
  mpz_t c;
  if (argc == 2 && isfile(argv[0], &fp) && ismpz(argv[1], c) &&
      mpz_sgn(c) >= 0 && my_mpz_fits_uint_p(c)) {
    mpz_t u;
    unsigned long count = mpz_get_ui(c), n = count/BYTES_PER_LIMB,
      k = count-n*BYTES_PER_LIMB, l = 0, i;
    if (k != 0) {
      n++;
#ifdef WORDS_BIGENDIAN
      l = BYTES_PER_LIMB-k;
#endif
    }
    if (!mpz_new(u, n)) return __ERROR;
    for (i = 0; i < n; i++) u->_mp_d[i] = 0;
    release_lock();
    if (read(fileno(fp), ((char*)u->_mp_d)+l, count) < count) {
      acquire_lock();
      mpz_clear(u);
      return __FAIL;
    }
    acquire_lock();
#ifdef WORDS_BIGENDIAN
    reverse_limbs(u->_mp_d, n);
#endif
    for (i = n; i > 0; i--)
      if (u->_mp_d[i-1] != 0) break;
    u->_mp_size = i;
    if (!mpz_actsize(u))
      return __FAIL;
    else
      return mkmpz(u);
  } else
    return __FAIL;
}

FUNCTION(clib,readfloat,argc,argv)
{
  FILE *fp;
  if (argc == 1 && isfile(argv[0], &fp)) {
    double f;
    release_lock();
    if (read(fileno(fp), &f, sizeof(double)) < sizeof(double)) {
      acquire_lock();
      return __FAIL;
    } else {
      acquire_lock();
      return mkfloat(f);
    }
  } else
    return __FAIL;
}

FUNCTION(clib,readcfloat,argc,argv)
{
  FILE *fp;
  if (argc == 1 && isfile(argv[0], &fp)) {
    float f;
    release_lock();
    if (read(fileno(fp), &f, sizeof(float)) < sizeof(float)) {
      acquire_lock();
      return __FAIL;
    } else {
      acquire_lock();
      return mkfloat((double)f);
    }
  } else
    return __FAIL;
}

FUNCTION(clib,readstr,argc,argv)
{
  FILE *fp;
  mpz_t c;
  if (argc == 2 && isfile(argv[0], &fp) && ismpz(argv[1], c) &&
      mpz_sgn(c) >= 0 && my_mpz_fits_uint_p(c)) {
    unsigned long count = mpz_get_ui(c);
    char *s = malloc(count+1);
    if (s) {
      release_lock();
      if (read(fileno(fp), s, count) == count) {
	acquire_lock();
	s[count] = 0;
	s = realloc(s, strlen(s)+1);
	return mkstr(s);
      } else {
	acquire_lock();
	free(s);
	return __FAIL;
      }
    } else
      return __ERROR;
  } else
    return __FAIL;
}

FUNCTION(clib,writeint,argc,argv)
{
  FILE *fp;
  mpz_t c, z;
  if (argc == 3 && isfile(argv[0], &fp) && ismpz(argv[1], c) &&
      mpz_sgn(c) >= 0 && my_mpz_fits_uint_p(c) && ismpz(argv[2], z)) {
    unsigned long count = mpz_get_ui(c), extra = 0, n = mpz_size(z),
      k = n*BYTES_PER_LIMB, i = 0;
    unsigned char pad_byte = 0;
    mpz_t u;
    mp_limb_t x;

    if (!mpz_new(u, n)) return __ERROR;
    mpz_set(u, z);
    /* determine the highest relevant byte */
    i = 0;
    if (n > 0)
      for (x = u->_mp_d[n-1]; x != 0 && i < BYTES_PER_LIMB; i++)
	x >>= 8;
    if (mpz_sgn(z) < 0) {
      unsigned long j;
      pad_byte = 0xff;
      /* 2's complement */
      for (j = 0; j < n; j++) u->_mp_d[j] = ~u->_mp_d[j];
      mpn_add_1(u->_mp_d, u->_mp_d, n, 1U);
    }
    if (n == 0) {
      extra = count;
      count = 0;
    } else if (count >= k) {
      extra = count-k;
      count = k;
#ifdef WORDS_BIGENDIAN
      reverse_limbs(u->_mp_d, n);
#endif
    }
#ifdef WORDS_BIGENDIAN
    else {
      /* adjust the value to the right */
      if ((n-1)*BYTES_PER_LIMB+i < count)
	i = count-(n-1)*BYTES_PER_LIMB;
      if (i < BYTES_PER_LIMB)
	mpn_lshift(u, u, n, (BYTES_PER_LIMB-i)<<3);
      /* reverse into big-endian limb order */
      reverse_limbs(u->_mp_d, n);
    }
    for (i = 0; i < extra; i++)
      if (write(fileno(fp), &pad_byte, 1) < 1)
	goto errexit;
#endif
    if (write(fileno(fp), u->_mp_d, count) < count)
      goto errexit;
    mpz_clear(u);
#ifndef WORDS_BIGENDIAN
    for (i = 0; i < extra; i++)
      if (write(fileno(fp), &pad_byte, 1) < 1)
	goto errexit;
#endif
    return mkvoid;
 errexit:
    mpz_clear(u);
    return __FAIL;
  } else
    return __FAIL;
}

FUNCTION(clib,writefloat,argc,argv)
{
  FILE *fp;
  double f;
  if (argc == 2 && isfile(argv[0], &fp) && isfloat(argv[1], &f)) {
    if (write(fileno(fp), &f, sizeof(double)) < sizeof(double))
      return __FAIL;
    else
      return mkvoid;
  } else
    return __FAIL;
}

FUNCTION(clib,writecfloat,argc,argv)
{
  FILE *fp;
  double f;
  if (argc == 2 && isfile(argv[0], &fp) && isfloat(argv[1], &f)) {
    float f1 = (float)f;
    if (write(fileno(fp), &f1, sizeof(float)) < sizeof(float))
      return __FAIL;
    else
      return mkvoid;
  } else
    return __FAIL;
}

/* C-style formatted I/O: *************************************************/

#define F_SZ 1024

static int f_err, f_var_wd, f_var_prec;
static int f_wd, f_prec;

static char *f_ptr, f_str[F_SZ], f_format[F_SZ], f_flags[F_SZ], f_mod[F_SZ];
static char f_wd_str[F_SZ], f_prec_str[F_SZ];

static void f_init(char *format)
{
  f_ptr = format;
  f_err = 0;
}

#define F_SZ_CHK(len) if (len>=F_SZ) { f_err = 1; return 0; } else

/* scan printf format string */

static char f_parse_pf(void)
{
  char c, *p, *q, *r;
  f_var_wd = f_var_prec = f_wd = f_prec = 0;
  for (p = strchr(f_ptr, '%'); p && p[1] == '%'; p = strchr(p+2, '%'))
    ;
  if (!p) {
    /* no more conversions, return the rest of the string */
    F_SZ_CHK(strlen(f_ptr)) strcpy(f_str, f_ptr);
    *f_format = *f_flags = *f_mod = 0;
    f_ptr = f_ptr + strlen(f_ptr);
    return 0;
  }
  /* conversion starts at p, parse it */
  F_SZ_CHK(p-f_ptr) strncpy(f_str, f_ptr, p-f_ptr);
  f_str[p-f_ptr] = 0;
  q = r = p+1;
  /* flags */
  while (strchr("#0- +", *r)) r++;
  F_SZ_CHK(r-q) strncpy(f_flags, q, r-q);
  f_flags[r-q] = 0;
  q = r;
  /* field width */
  if (*r == '*') {
    r++;
    f_var_wd = 1;
    *f_wd_str = 0;
  } else {
    while (isdigit(*r)) r++;
    F_SZ_CHK(r-q) strncpy(f_wd_str, q, r-q);
    f_wd_str[r-q] = 0;
  }
  if (*f_wd_str) f_wd = atoi(f_wd_str);
  q = r;
  /* precision */
  if (*r == '.') {
    r++;
    if (*r == '*') {
      r++;
      f_var_prec = 1;
      *f_prec_str = 0;
    } else {
      while (isdigit(*r)) r++;
      F_SZ_CHK(r-q) strncpy(f_prec_str, q, r-q);
      f_prec_str[r-q] = 0;
    }
    if (*f_prec_str) f_prec = atoi(f_prec_str);
  }
  q = r;
  /* length modifier */
  while (strchr("hl", *r)) r++;
  F_SZ_CHK(r-q) strncpy(f_mod, q, r-q);
  f_mod[r-q] = 0;
  q = r;
  /* conversion specifier */
  c = *(r++);
  F_SZ_CHK(r-p) strncpy(f_format, p, r-p);
  f_format[r-p] = 0;
  f_ptr = r;
  return c;
}

/* scan scanf format string */

static char f_parse_sf(void)
{
  char c, *p, *q, *r;
  f_wd = -1;
  for (p = strchr(f_ptr, '%'); p && p[1] == '%'; p = strchr(p+2, '%'))
    ;
  if (!p) {
    /* no more conversions, return the rest of the string */
    F_SZ_CHK(strlen(f_ptr)) strcpy(f_str, f_ptr);
    *f_format = *f_flags = *f_mod = 0;
    f_ptr = f_ptr + strlen(f_ptr);
    return 0;
  }
  /* conversion starts at p, parse it */
  F_SZ_CHK(p-f_ptr) strncpy(f_str, f_ptr, p-f_ptr);
  f_str[p-f_ptr] = 0;
  q = r = p+1;
  /* flags */
  while (strchr("*", *r)) r++;
  F_SZ_CHK(r-q) strncpy(f_flags, q, r-q);
  f_flags[r-q] = 0;
  q = r;
  /* field width */
  while (isdigit(*r)) r++;
  F_SZ_CHK(r-q) strncpy(f_wd_str, q, r-q);
  f_wd_str[r-q] = 0;
  if (*f_wd_str) f_wd = atoi(f_wd_str);
  q = r;
  /* length modifier */
  while (strchr("hl", *r)) r++;
  F_SZ_CHK(r-q) strncpy(f_mod, q, r-q);
  f_mod[r-q] = 0;
  q = r;
  /* conversion specifier */
  c = *(r++);
  if (c == '[') {
    /* character class */
    if (*r == ']')
      r++;
    else if (r[0] == '^' && r[1] == ']')
      r += 2;
    while (*r && *r != ']') r++;
    if (*r == ']')
      r++;
    else {
      f_err = 1;
      c = 0;
    }
  }
  F_SZ_CHK(r-p) strncpy(f_format, p, r-p);
  f_format[r-p] = 0;
  f_ptr = r;
  return c;
}

static int ismpz_int(expr x, unsigned long *u)
{
  mpz_t z;
  if (ismpz(x, z)) {
    if (mpz_sgn(z) != 0)
      *u = z->_mp_d[0];
    else
      *u = 0;
    if (mpz_sgn(z) < 0)
      *u = ~(*u)+1U;
    return 1;
  } else
    return 0;
}

#define CALL_FPRINTF(fp,val) \
((f_var_wd && f_var_prec)? \
fprintf(fp, f_format, (int)wdval, (int)precval, val): \
(f_var_wd)? \
fprintf(fp, f_format, (int)wdval, val): \
(f_var_prec)? \
fprintf(fp, f_format, (int)precval, val): \
fprintf(fp, f_format, val))

FUNCTION(clib,fprintf,argc,argv)
{
  expr x, *xs;
  int i, n, ret;
  char f, *format;
  FILE *fp;

  long intval, wdval, precval;
  unsigned long uintval;
  double dblval;
  char *strval;

  if (argc != 3 || !isfile(argv[0], &fp) || !isstr(argv[1], &format))
    return __FAIL;
  lock_format();
  if (!istuple(argv[2], &n, &xs)) {
    x = argv[2];
    xs = &x;
    n = 1;
  }
  /* parse the format string and check arguments */
  f_init(format);
  i = 0;
  while ((f = f_parse_pf())) {
    if (*f_mod)
      if (strcmp(f_mod, "l") && strcmp(f_mod, "h") ||
	  !strchr("diouxX", f)) {
	unlock_format();
	return __FAIL;
      }
    if (f_var_wd)
      if ((i >= n || !ismpz_int(xs[i], &wdval))) {
	unlock_format();
	return __FAIL;
      } else
	i++;
    if (f_var_prec)
      if ((i >= n || !ismpz_int(xs[i], &precval))) {
	unlock_format();
	return __FAIL;
      } else
	i++;
    switch (f) {
    case 'd': case 'i':
      if (i >= n || !ismpz_int(xs[i], &intval)) {
	unlock_format();
	return __FAIL;
      } else
	i++;
      break;
    case 'o': case 'u': case 'x': case 'X':
      if (i >= n || !ismpz_int(xs[i], &uintval)) {
	unlock_format();
	return __FAIL;
      } else
	i++;
      break;
    case 'e': case 'E':
    case 'f':
    case 'g': case 'G':
      if (i >= n || !isfloat(xs[i], &dblval)) {
	unlock_format();
	return __FAIL;
      } else
	i++;
      break;
    case 'c':
      if (i >= n || !isstr(xs[i], &strval) || !strval[0] || strval[1]) {
	unlock_format();
	return __FAIL;
      } else
	i++;
      break;
    case 's':
      if (i >= n || !isstr(xs[i], &strval)) {
	unlock_format();
	return __FAIL;
      } else
	i++;
      break;
    default:
      unlock_format();
      return __FAIL;
    }
  }
  if (f_err) {
    unlock_format();
    return __FAIL;
  }
  /* print */
  f_init(format);
  i = 0;
  while ((f = f_parse_pf())) {
    if (fprintf(fp, f_str) < 0) goto errexit;
    if (f_var_wd)
      ismpz_int(xs[i++], &wdval);
    if (f_var_prec)
      ismpz_int(xs[i++], &precval);
    ret = 0;
    switch (f) {
    case 'd': case 'i':
      ismpz_int(xs[i++], &intval);
      if (*f_mod == 'l')
	ret = CALL_FPRINTF(fp, intval);
      else if (*f_mod == 'h')
	ret = CALL_FPRINTF(fp, (short)intval);
      else
	ret = CALL_FPRINTF(fp, (int)intval);
      break;
    case 'o': case 'u': case 'x': case 'X':
      ismpz_int(xs[i++], &uintval);
      if (*f_mod == 'l')
	ret = CALL_FPRINTF(fp, uintval);
      else if (*f_mod == 'h')
	ret = CALL_FPRINTF(fp, (unsigned short)uintval);
      else
	ret = CALL_FPRINTF(fp, (unsigned)uintval);
      break;
    case 'e': case 'E':
    case 'f':
    case 'g': case 'G':
      isfloat(xs[i++], &dblval);
      ret = CALL_FPRINTF(fp, dblval);
      break;
    case 'c':
      isstr(xs[i++], &strval);
      ret = CALL_FPRINTF(fp, (int)*strval);
      break;
    case 's':
      isstr(xs[i++], &strval);
      ret = CALL_FPRINTF(fp, strval);
      break;
    }
    if (ret < 0) goto errexit;
  }
  if (fprintf(fp, f_str) < 0) goto errexit;
  f_init(NULL);
  unlock_format();
  return mkvoid;
 errexit:
  f_init(NULL);
  unlock_format();
  return __FAIL;
}

FUNCTION(clib,printf,argc,argv)
{
  char *format;
  if (argc == 2 && isstr(argv[0], &format)) {
    expr output = eval(mksym(sym(OUTPUT)));
    if (output) {
      expr argv1[3] = { output, argv[0], argv[1] };
      expr ret = FUNCALL(clib, fprintf, 3, argv1);
      dispose(output);
      return ret;
    } else
      return __FAIL;
  } else
    return __FAIL;
}

static char *buf = NULL, *bufptr = NULL;
static long alloc = 0, leng = 0;

static int addbuf(long newleng)
{
  if (!buf)
    alloc = leng = 0;
  else
    leng = strlen(buf);
  if (leng+newleng+1 <= 0) {
    if (!buf) bufptr = NULL;
    return 0;
  }
  while (leng+newleng >= alloc) {
    if (buf) {
      char *newbuf;
      if (alloc+BUFSZ <= 0 || !(newbuf = realloc(buf, alloc+BUFSZ))) {
	bufptr = buf+leng;
	return 0;
      } else {
	buf = newbuf;
	alloc += BUFSZ;
      }
    } else {
      if (!(buf = malloc(BUFSZ))) {
	bufptr = NULL;
	return 0;
      } else {
	alloc = BUFSZ;
	*buf = 0;
      }
    }
  }
  bufptr = buf+leng;
  return 1;
}

#define CALL_SPRINTF(s,val) \
((f_var_wd && f_var_prec)? \
sprintf(s, f_format, (int)wdval, (int)precval, val): \
(f_var_wd)? \
sprintf(s, f_format, (int)wdval, val): \
(f_var_prec)? \
sprintf(s, f_format, (int)precval, val): \
sprintf(s, f_format, val))

FUNCTION(clib,sprintf,argc,argv)
{
  expr x, *xs;
  int i, n, ret;
  long sz, k, l;
  char f, *format, *s;

  long intval, wdval, precval;
  unsigned long uintval;
  double dblval;
  char *strval;

  if (argc != 2 || !isstr(argv[0], &format))
    return __FAIL;
  lock_format();
  if (!istuple(argv[1], &n, &xs)) {
    x = argv[1];
    xs = &x;
    n = 1;
  }
  /* parse the format string and check arguments, guestimate needed
     buffer size */
  f_init(format);
  i = 0; sz = BUFSZ;
  while ((f = f_parse_pf())) {
    if (*f_mod)
      if (strcmp(f_mod, "l") && strcmp(f_mod, "h") ||
	  !strchr("diouxX", f)) {
	unlock_format();
	return __FAIL;
      }
    if (f_var_wd)
      if ((i >= n || !ismpz_int(xs[i], &wdval))) {
	unlock_format();
	return __FAIL;
      } else {
	f_wd = wdval;
	i++;
      }
    if (f_var_prec)
      if ((i >= n || !ismpz_int(xs[i], &precval))) {
	unlock_format();
	return __FAIL;
      } else {
	f_prec = precval;
	i++;
      }
    if (f_wd < 0) f_wd = -f_wd;
    if (f_prec < 0) f_prec = 0;
    l = f_wd+f_prec;
    switch (f) {
    case 'd': case 'i':
      if (i >= n || !ismpz_int(xs[i], &intval)) {
	unlock_format();
	return __FAIL;
      } else
	i++;
      if (BUFSZ+l > sz) sz = BUFSZ+l;
      break;
    case 'o': case 'u': case 'x': case 'X':
      if (i >= n || !ismpz_int(xs[i], &uintval)) {
	unlock_format();
	return __FAIL;
      } else
	i++;
      if (BUFSZ+l > sz) sz = BUFSZ+l;
      break;
    case 'e': case 'E':
    case 'f':
    case 'g': case 'G':
      if (i >= n || !isfloat(xs[i], &dblval)) {
	unlock_format();
	return __FAIL;
      } else
	i++;
      if (f == 'f') {
	k = log10(abs(dblval)+1)+2;
	if (BUFSZ+l+k > sz) sz = BUFSZ+l+k;
      } else {
	if (BUFSZ+l > sz) sz = BUFSZ+l;
      }
      break;
    case 'c':
      if (i >= n || !isstr(xs[i], &strval) || !strval[0] || strval[1]) {
	unlock_format();
	return __FAIL;
      } else
	i++;
      if (l+1 > sz) sz = l+1;
      break;
    case 's':
      if (i >= n || !isstr(xs[i], &strval)) {
	unlock_format();
	return __FAIL;
      } else
	i++;
      k = strlen(strval);
      if (k+l > sz) sz = k+l;
      break;
    default:
      unlock_format();
      return __FAIL;
    }
  }
  if (f_err) {
    unlock_format();
    return __FAIL;
  }
  /* allocate buffer */
  if (sz <= 0 || !(s = malloc(sz))) {
    unlock_format();
    return __ERROR;
  }
  buf = NULL;
  /* print */
  f_init(format);
  i = 0;
  while ((f = f_parse_pf())) {
    if (!addbuf(strlen(f_str))) goto errexit2;
    if (sprintf(bufptr, f_str) < 0) goto errexit;
    bufptr += strlen(bufptr);
    if (f_var_wd)
      ismpz_int(xs[i++], &wdval);
    if (f_var_prec)
      ismpz_int(xs[i++], &precval);
    *s = 0;
    switch (f) {
    case 'd': case 'i':
      ismpz_int(xs[i++], &intval);
      if (*f_mod == 'l')
	ret = CALL_SPRINTF(s, intval);
      else if (*f_mod == 'h')
	ret = CALL_SPRINTF(s, (short)intval);
      else
	ret = CALL_SPRINTF(s, (int)intval);
      break;
    case 'o': case 'u': case 'x': case 'X':
      ismpz_int(xs[i++], &uintval);
      if (*f_mod == 'l')
	ret = CALL_SPRINTF(s, uintval);
      else if (*f_mod == 'h')
	ret = CALL_SPRINTF(s, (unsigned short)uintval);
      else
	ret = CALL_SPRINTF(s, (unsigned)uintval);
      break;
    case 'e': case 'E':
    case 'f':
    case 'g': case 'G':
      isfloat(xs[i++], &dblval);
      ret = CALL_SPRINTF(s, dblval);
      break;
    case 'c':
      isstr(xs[i++], &strval);
      ret = CALL_SPRINTF(s, (int)*strval);
      break;
    case 's':
      isstr(xs[i++], &strval);
      ret = CALL_SPRINTF(s, strval);
      break;
    }
    if (ret < 0) goto errexit;
    if (!addbuf(strlen(s))) goto errexit2;
    strcpy(bufptr, s);
    bufptr += strlen(bufptr);
  }
  if (!addbuf(strlen(f_str))) goto errexit2;
  if (sprintf(bufptr, f_str) < 0) goto errexit;
  bufptr += strlen(bufptr);
  f_init(NULL);
  unlock_format();
  free(s);
  buf = realloc(buf, strlen(buf)+1);
  return mkstr(buf);
 errexit:
  f_init(NULL);
  unlock_format();
  if (buf) free(buf); free(s);
  return __FAIL;
 errexit2:
  f_init(NULL);
  unlock_format();
  if (buf) free(buf); free(s);
  return __ERROR;
}

#define DFLT_STRLEN 10240

/* scanning literals is a bit tricky, since there is no obvious sign
   indicating that the match has failed -- we have to check for non-ws chars
   in the template */
static int fscan_literal(FILE *fp, long *ncount, char *template)
{
  long count = -1;
  short empty = 1;
  char *p;

  for (p = template; *p; p++)
    if (!isspace(*p)) {
      empty = 0;
      break;
    }
  if (strlen(template) >= F_SZ-3) return 0;
  strcat(template, "%ln");
  if (fscanf(fp, template, &count) < 0) return 0;
  if (count < 0) return 0;
  if (count == 0 && !empty) return 0;
  *ncount += count;
  return 1;
}

FUNCTION(clib,fscanf,argc,argv)
{
  expr *xs, x;
  int i, n, ret;
  long count, ncount = 0;
  char f, *format;
  FILE *fp;

  short shortval;
  unsigned short ushortval;
  int intval;
  unsigned uintval;
  long longval;
  unsigned long ulongval;
  float fltval;
  double dblval;
  char *strval;

  if (argc != 2 || !isfile(argv[0], &fp) || !isstr(argv[1], &format))
    return __FAIL;
  release_lock();
  if (fp == stdin) acquire_tty();
  lock_format();
  f_init(format);
  i = 0;
  while ((f = f_parse_sf())) {
    if (*f_flags)
      if (strcmp(f_flags, "*")) {
	unlock_format();
	if (fp == stdin) release_tty();
	acquire_lock();
	return __FAIL;
      }
    if (*f_mod)
      if (strcmp(f_mod, "l") && strcmp(f_mod, "h") ||
	  *f_mod == 'l' && !strchr("ndiouxXeEfgG", f) ||
	  *f_mod == 'h' && !strchr("ndiouxX", f)) {
	unlock_format();
	if (fp == stdin) release_tty();
	acquire_lock();
	return __FAIL;
      }
    switch (f) {
    case 'n':
    case 'd': case 'i':
    case 'o': case 'u': case 'x': case 'X':
    case 'e': case 'E':
    case 'f':
    case 'g': case 'G':
    case 'c':
    case 's': case '[':
      if (!*f_flags) i++;
      break;
    default:
      unlock_format();
      if (fp == stdin) release_tty();
      acquire_lock();
      return __FAIL;
    }
  }
  if (f_err) {
    unlock_format();
    if (fp == stdin) release_tty();
    acquire_lock();
    return __FAIL;
  }
  n = i;
  if (n == 0)
    xs = NULL;
  else if (!(xs = malloc(n*sizeof(expr)))) {
    unlock_format();
    if (fp == stdin) release_tty();
    acquire_lock();
    return __ERROR;
  }
  f_init(format);
  i = 0;
  while ((f = f_parse_sf())) {
    if (!fscan_literal(fp, &ncount, f_str)) goto errexit;
    if (strlen(f_format) >= F_SZ-3) goto errexit;
    strcat(f_format, "%ln");
    if (*f_flags) {
      count = -1; fscanf(fp, f_format, &count);
      if (count < 0) goto errexit;
      ncount += count;
      continue;
    }
    switch (f) {
    case 'n':
      if (*f_mod == 'h') {
	shortval = ncount;
	ret = (xs[i] = mkint(shortval)) != NULL;
      } else if (*f_mod == 'h') {
	longval = ncount;
	ret = (xs[i] = mkint(longval)) != NULL;
      } else {
	intval = ncount;
	ret = (xs[i] = mkint(intval)) != NULL;
      }
      if (!ret)
	goto errexit2;
      else
	i++;
      count = 0;
      break;
    case 'd': case 'i':
      count = -1;
      if (*f_mod == 'h')
	ret = !(fscanf(fp, f_format, &shortval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkint(shortval))?-1:0;
      else if (*f_mod == 'l')
	ret = !(fscanf(fp, f_format, &longval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkint(longval))?-1:0;
      else
	ret = !(fscanf(fp, f_format, &intval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkint(intval))?-1:0;
      if (ret == -1)
	goto errexit2;
      else if (ret == 1)
	goto errexit;
      else
	i++;
      break;
    case 'o': case 'u': case 'x': case 'X':
      count = -1;
      if (*f_mod == 'h')
	ret = !(fscanf(fp, f_format, &ushortval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkuint(ushortval))?-1:0;
      else if (*f_mod == 'l')
	ret = !(fscanf(fp, f_format, &ulongval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkuint(ulongval))?-1:0;
      else
	ret = !(fscanf(fp, f_format, &uintval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkuint(uintval))?-1:0;
      if (ret == -1)
	goto errexit2;
      else if (ret == 1)
	goto errexit;
      else
	i++;
      break;
    case 'e': case 'E':
    case 'f':
    case 'g': case 'G':
      count = -1;
      if (*f_mod == 'l')
	ret = !(fscanf(fp, f_format, &dblval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkfloat(dblval))?-1:0;
      else
	ret = !(fscanf(fp, f_format, &fltval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkfloat(fltval))?-1:0;
      if (ret == -1)
	goto errexit2;
      else if (ret == 1)
	goto errexit;
      else
	i++;
      break;
    case 'c':
      count = -1;
      if (f_wd <= 0) f_wd = 1;
      if (!(strval = malloc(f_wd+1))) goto errexit2;
      *strval = 0;
      if (fscanf(fp, f_format, strval, &count) <= 0 || count < 0) {
	free(strval);
	goto errexit;
      } else {
	strval[f_wd] = 0;
	strval = realloc(strval, strlen(strval)+1);
	if (!(xs[i] = mkstr(strval)))
	  goto errexit2;
	else
	  i++;
      }
      break;
    case 's': case '[':
      count = -1;
      if (f_wd <= 0) {
	char new_format[F_SZ];
	f_wd = DFLT_STRLEN;
	if (strlen(f_format)+10 > F_SZ) goto errexit;
	sprintf(new_format, "%%%d%s", f_wd, f_format+1);
	strcpy(f_format, new_format);
      }
      if (!(strval = malloc(f_wd+1))) goto errexit2;
      *strval = 0;
      if (fscanf(fp, f_format, strval, &count) <= 0 || count < 0) {
	free(strval);
	goto errexit;
      } else {
	strval = realloc(strval, strlen(strval)+1);
	if (!(xs[i] = mkstr(strval)))
	  goto errexit2;
	else
	  i++;
      }
      break;
    }
    ncount += count;
  }
  if (!fscan_literal(fp, &ncount, f_str)) goto errexit;
  unlock_format();
  if (fp == stdin) release_tty();
  acquire_lock();
  if (n == 0)
    return mkvoid;
  else if (n == 1) {
    x = xs[0]; free(xs);
    return x;
  } else
    return mktuplev(n, xs);
 errexit:
  unlock_format();
  if (fp == stdin) release_tty();
  acquire_lock();
  while (i > 0) dispose(xs[--i]);
  if (n > 0) free(xs);
  return __FAIL;
 errexit2:
  unlock_format();
  if (fp == stdin) release_tty();
  acquire_lock();
  while (i > 0) dispose(xs[--i]);
  if (n > 0) free(xs);
  return __ERROR;
}

FUNCTION(clib,scanf,argc,argv)
{
  char *format;
  if (argc == 1 && isstr(argv[0], &format)) {
    expr input = eval(mksym(sym(INPUT)));
    if (input) {
      expr argv1[2] = { input, argv[0] };
      expr ret = FUNCALL(clib, fscanf, 2, argv1);
      dispose(input);
      return ret;
    } else
      return __FAIL;
  } else
    return __FAIL;
}

static int sscan_literal(char **s, long *ncount, char *template)
{
  long count = -1;
  int empty = 1;
  char *p;

  for (p = template; *p; p++)
    if (!isspace(*p)) {
      empty = 0;
      break;
    }
  if (strlen(template) >= F_SZ-3) return 0;
  strcat(template, "%ln");
  if (sscanf(*s, template, &count) < 0) return 0;
  if (count < 0) return 0;
  if (count == 0 && !empty) return 0;
  *s += count;
  *ncount += count;
  return 1;
}

FUNCTION(clib,sscanf,argc,argv)
{
  expr *xs, x;
  int i, n, ret;
  long count, ncount = 0;
  char f, *format;
  char *s;

  short shortval;
  unsigned short ushortval;
  int intval;
  unsigned uintval;
  long longval;
  unsigned long ulongval;
  float fltval;
  double dblval;
  char *strval;

  if (argc != 2 || !isstr(argv[0], &s) || !isstr(argv[1], &format))
    return __FAIL;
  release_lock();
  lock_format();
  f_init(format);
  i = 0;
  while ((f = f_parse_sf())) {
    if (*f_flags)
      if (strcmp(f_flags, "*")) {
	unlock_format();
	acquire_lock();
	return __FAIL;
      }
    if (*f_mod)
      if (strcmp(f_mod, "l") && strcmp(f_mod, "h") ||
	  *f_mod == 'l' && !strchr("ndiouxXeEfgG", f) ||
	  *f_mod == 'h' && !strchr("ndiouxX", f)) {
	unlock_format();
	acquire_lock();
	return __FAIL;
      }
    switch (f) {
    case 'n':
    case 'd': case 'i':
    case 'o': case 'u': case 'x': case 'X':
    case 'e': case 'E':
    case 'f':
    case 'g': case 'G':
    case 'c':
    case 's': case '[':
      if (!*f_flags) i++;
      break;
    default:
      unlock_format();
      acquire_lock();
      return __FAIL;
    }
  }
  if (f_err) {
    unlock_format();
    acquire_lock();
    return __FAIL;
  }
  n = i;
  if (n == 0)
    xs = NULL;
  else if (!(xs = malloc(n*sizeof(expr)))) {
    unlock_format();
    acquire_lock();
    return __ERROR;
  }
  f_init(format);
  i = 0;
  while ((f = f_parse_sf())) {
    if (!sscan_literal(&s, &ncount, f_str)) goto errexit;
    if (strlen(f_format) >= F_SZ-3) goto errexit;
    strcat(f_format, "%ln");
    if (*f_flags) {
      count = -1; sscanf(s, f_format, &count);
      if (count < 0) goto errexit;
      s += count; ncount += count;
      continue;
    }
    switch (f) {
    case 'n':
      if (*f_mod == 'h') {
	shortval = ncount;
	ret = (xs[i] = mkint(shortval)) != NULL;
      } else if (*f_mod == 'h') {
	longval = ncount;
	ret = (xs[i] = mkint(longval)) != NULL;
      } else {
	intval = ncount;
	ret = (xs[i] = mkint(intval)) != NULL;
      }
      if (!ret)
	goto errexit2;
      else
	i++;
      count = 0;
      break;
    case 'd': case 'i':
      count = -1;
      if (*f_mod == 'h')
	ret = !(sscanf(s, f_format, &shortval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkint(shortval))?-1:0;
      else if (*f_mod == 'l')
	ret = !(sscanf(s, f_format, &longval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkint(longval))?-1:0;
      else
	ret = !(sscanf(s, f_format, &intval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkint(intval))?-1:0;
      if (ret == -1)
	goto errexit2;
      else if (ret == 1)
	goto errexit;
      else
	i++;
      break;
    case 'o': case 'u': case 'x': case 'X':
      count = -1;
      if (*f_mod == 'h')
	ret = !(sscanf(s, f_format, &ushortval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkuint(ushortval))?-1:0;
      else if (*f_mod == 'l')
	ret = !(sscanf(s, f_format, &ulongval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkuint(ulongval))?-1:0;
      else
	ret = !(sscanf(s, f_format, &uintval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkuint(uintval))?-1:0;
      if (ret == -1)
	goto errexit2;
      else if (ret == 1)
	goto errexit;
      else
	i++;
      break;
    case 'e': case 'E':
    case 'f':
    case 'g': case 'G':
      count = -1;
      if (*f_mod == 'l')
	ret = !(sscanf(s, f_format, &dblval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkfloat(dblval))?-1:0;
      else
	ret = !(sscanf(s, f_format, &fltval, &count) > 0 && count >= 0)?1:
	  !(xs[i] = mkfloat(fltval))?-1:0;
      if (ret == -1)
	goto errexit2;
      else if (ret == 1)
	goto errexit;
      else
	i++;
      break;
    case 'c':
      count = -1;
      if (f_wd == -1) f_wd = 1;
      if (f_wd > strlen(s)) goto errexit;
      if (!(strval = malloc(f_wd+1))) goto errexit2;
      *strval = 0;
      if (sscanf(s, f_format, strval, &count) <= 0 || count < 0) {
	free(strval);
	goto errexit;
      } else {
	strval[f_wd] = 0;
	strval = realloc(strval, strlen(strval)+1);
	if (!(xs[i] = mkstr(strval)))
	  goto errexit2;
	else
	  i++;
      }
      break;
    case 's': case '[':
      count = -1;
      if (f_wd == -1) {
	char new_format[F_SZ];
	f_wd = DFLT_STRLEN;
	if (strlen(f_format)+10 > F_SZ) goto errexit;
	sprintf(new_format, "%%%d%s", f_wd, f_format+1);
	strcpy(f_format, new_format);
      }
      if (!(strval = malloc(f_wd+1))) goto errexit2;
      *strval = 0;
      if (sscanf(s, f_format, strval, &count) <= 0 || count < 0) {
	free(strval);
	goto errexit;
      } else {
	strval = realloc(strval, strlen(strval)+1);
	if (!(xs[i] = mkstr(strval)))
	  goto errexit2;
	else
	  i++;
      }
      break;
    default:
      goto errexit;
    }
    s += count; ncount += count;
  }
  if (!sscan_literal(&s, &ncount, f_str)) goto errexit;
  unlock_format();
  acquire_lock();
  if (n == 0)
    return mkvoid;
  else if (n == 1) {
    x = xs[0]; free(xs);
    return x;
  } else
    return mktuplev(n, xs);
 errexit:
  unlock_format();
  acquire_lock();
  while (i > 0) dispose(xs[--i]);
  if (n > 0) free(xs);
  return __FAIL;
 errexit2:
  unlock_format();
  acquire_lock();
  while (i > 0) dispose(xs[--i]);
  if (n > 0) free(xs);
  return __ERROR;
}

/* file and system utilities: *********************************************/

FUNCTION(clib,rename,argc,argv)
{
  char *old, *new;
  if (argc == 2 && isstr(argv[0], &old) && isstr(argv[1], &new))
    if (rename(old, new))
      return __FAIL;
    else
      return mkvoid;
  else
    return __FAIL;
}

FUNCTION(clib,unlink,argc,argv)
{
  char *name;
  if (argc == 1 && isstr(argv[0], &name))
    if (unlink(name))
      return __FAIL;
    else
      return mkvoid;
  else
    return __FAIL;
}

FUNCTION(clib,getcwd,argc,argv)
{
  char pwd[BUFSZ];
  if (argc == 0 && getcwd(pwd, BUFSZ)) {
    char *s = strdup(pwd);
    if (s)
      return mkstr(s);
    else
      return __ERROR;
  } else
    return __FAIL;
}

FUNCTION(clib,chdir,argc,argv)
{
  char *name;
  if (argc == 1 && isstr(argv[0], &name))
    if (chdir(name))
      return __FAIL;
    else
      return mkvoid;
  else
    return __FAIL;
}

FUNCTION(clib,mkdir,argc,argv)
{
  char *name;
  if (argc == 1 && isstr(argv[0], &name))
    if (mkdir(name, 0777))
      return __FAIL;
    else
      return mkvoid;
  else
    return __FAIL;
}

FUNCTION(clib,rmdir,argc,argv)
{
  char *name;
  if (argc == 1 && isstr(argv[0], &name))
    if (rmdir(name))
      return __FAIL;
    else
      return mkvoid;
  else
    return __FAIL;
}

FUNCTION(clib,mkfifo,argc,argv)
{
#ifdef __MINGW32__
  return __FAIL;
#else
  char *name;
  if (argc == 1 && isstr(argv[0], &name))
    if (mkfifo(name, 0666))
      return __FAIL;
    else
      return mkvoid;
  else
    return __FAIL;
#endif
}

FUNCTION(clib,umask,argc,argv)
{
  long mask;
  if (argc == 1 && isint(argv[0], &mask) && mask >= 0 && mask <= 0777)
    return mkint(umask(mask));
  else
    return __FAIL;
}

FUNCTION(clib,stat,argc,argv)
{
  char *name;
  if (argc == 1 && isstr(argv[0], &name)) {
    struct stat buf;
    expr *st, x;
    if (stat(name, &buf))
      return __FAIL;
    else {
      if (!(st = malloc(11*sizeof(expr)))) return __ERROR;
      /* FIXME: all types are assumed to fit into a 4 byte value; this may
	 not be true for some (dev_t, maybe others?) */
      st[0] = mkuint(buf.st_dev);
      st[1] = mkuint(buf.st_ino);
      st[2] = mkuint(buf.st_mode);
      st[3] = mkuint(buf.st_nlink);
      st[4] = mkuint(buf.st_uid);
      st[5] = mkuint(buf.st_gid);
#ifdef HAVE_ST_RDEV
      st[6] = mkuint(buf.st_rdev);
#else
      st[6] = mkuint(0);
#endif
      st[7] = mkint(buf.st_size);
      st[8] = mkint(buf.st_atime);
      st[9] = mkint(buf.st_mtime);
      st[10] = mkint(buf.st_ctime);
      x = mktuplev(11, st);
      if (x)
	return x;
      else
	return __ERROR;
    }
  } else
    return __FAIL;
}

#ifdef WIN32
#define lstat stat
#endif

FUNCTION(clib,lstat,argc,argv)
{
  char *name;
  if (argc == 1 && isstr(argv[0], &name)) {
    struct stat buf;
    expr *st, x;
    if (lstat(name, &buf))
      return __FAIL;
    else {
      if (!(st = malloc(11*sizeof(expr)))) return __ERROR;
      /* FIXME: all types are assumed to fit into a 4 byte value; this may
	 not be true for some (dev_t, maybe others?) */
      st[0] = mkuint(buf.st_dev);
      st[1] = mkuint(buf.st_ino);
      st[2] = mkuint(buf.st_mode);
      st[3] = mkuint(buf.st_nlink);
      st[4] = mkuint(buf.st_uid);
      st[5] = mkuint(buf.st_gid);
      st[6] = mkuint(buf.st_rdev);
      st[7] = mkint(buf.st_size);
      st[8] = mkint(buf.st_atime);
      st[9] = mkint(buf.st_mtime);
      st[10] = mkint(buf.st_ctime);
      x = mktuplev(11, st);
      if (x)
	return x;
      else
	return __ERROR;
    }
  } else
    return __FAIL;
}

FUNCTION(clib,system,argc,argv)
{
  char *cmd;
  if (argc == 1 && isstr(argv[0], &cmd)) {
    int ret;
    errno = 0;
    if ((ret = system(cmd)))
      if (errno)
	return __FAIL;
      else
	mkint(ret);
    else
      return mkint(ret);
  } else
    return __FAIL;
}

FUNCTION(clib,fork,argc,argv)
{
#ifdef __MINGW32__
  return __FAIL;
#else
  if (argc == 0/* && this_thread() == 0*/) {
    int pid = fork();
    if (pid < 0)
      return __FAIL;
    else
      return mkint(pid);
  } else
    return __FAIL;
#endif
}

FUNCTION(clib,exec,argc,argv)
{
  char *prog, **args;
  if (argc == 2 && isstr(argv[0], &prog)) {
    int n;
    expr x, hd, tl;
    char *s;
    for (n = 0, x = argv[1]; iscons(x, &hd, &tl); n++)
      if (n >= INT_MAX)
	return __ERROR;
      else if (!isstr(hd, &s))
	return __FAIL;
      else
	x = tl;
    if (!isnil(x)) return __FAIL;
    if (!(args = malloc((n+1)*sizeof(char*))))
      return __ERROR;
    for (n = 0, x = argv[1]; iscons(x, &hd, &tl); n++) {
      isstr(hd, args+n);
      x = tl;
    }
    args[n] = NULL;
    execvp(prog, args);
    /* if we come here, something has gone wrong */
    return __FAIL;
  } else
    return __FAIL;
}

static int spawn(char *prog, char *argv[], int *status)
{
#ifdef __MINGW32__
  *status = spawnvp(P_WAIT, prog, argv);
  return *status >= 0;
#else
  int pid;
  switch ((pid = fork())) {
  case 0:
    execvp(prog, argv);
  case -1:
    return 0;
  }
  waitpid(pid, status, 0);
  return 1;
#endif
}

FUNCTION(clib,spawn,argc,argv)
{
  char *prog, **args;
  if (argc == 2 && isstr(argv[0], &prog)) {
    int n, status;
    expr x, hd, tl;
    char *s;
    for (n = 0, x = argv[1]; iscons(x, &hd, &tl); n++)
      if (n >= INT_MAX)
	return __ERROR;
      else if (!isstr(hd, &s))
	return __FAIL;
      else
	x = tl;
    if (!isnil(x)) return __FAIL;
    if (!(args = malloc((n+1)*sizeof(char*))))
      return __ERROR;
    for (n = 0, x = argv[1]; iscons(x, &hd, &tl); n++) {
      isstr(hd, args+n);
      x = tl;
    }
    args[n] = NULL;
    if (spawn(prog, args, &status))
      return mkint(status);
    else
      return __FAIL;
  } else
    return __FAIL;
}

FUNCTION(clib,exit,argc,argv)
{
  long code;
  if (argc == 1 && isint(argv[0], &code)) {
    exit(code);
    /* we shouldn't arrive here */
    return __FAIL;
  } else
    return __FAIL;
}

FUNCTION(clib,kill,argc,argv)
{
#ifdef __MINGW32__
  return __FAIL;
#else
  long sig, pid;
  if (argc == 2 && isint(argv[0], &sig) && isint(argv[1], &pid)) {
    if (kill(pid, sig) < 0)
      return __FAIL;
    else
      return mkvoid;
  } else
    return __FAIL;
#endif
}

FUNCTION(clib,getpid,argc,argv)
{
  if (argc == 0)
    return mkint(getpid());
  else
    return __FAIL;
}

#ifndef __MINGW32__
static expr waitres(int res, int status)
{
  if (res < 0)
    return __FAIL;
  else if (res == 0)
    return mkvoid;
  else
    return mkint(status);
}
#endif

FUNCTION(clib,wait,argc,argv)
{
#ifdef __MINGW32__
  return __FAIL;
#else
  if (argc == 0) {
    int status, pid = wait(&status);
    return waitres(pid, status);
  } else
    return __FAIL;
#endif
}

FUNCTION(clib,waitpid,argc,argv)
{
#ifdef __MINGW32__
  return __FAIL;
#else
  long pid;
  if (argc == 1 && isint(argv[0], &pid)) {
    int status;
    pid = waitpid(pid, &status, 0);
    return waitres(pid, status);
  } else
    return __FAIL;
#endif
}

FUNCTION(clib,status,argc,argv)
{
#ifdef __MINGW32__
  return __FAIL;
#else
  long pid;
  if (argc == 1 && isint(argv[0], &pid)) {
    int status;
    pid = waitpid(pid, &status, WNOHANG|WUNTRACED);
    return waitres(pid, status);
  } else
    return __FAIL;
#endif
}

FUNCTION(clib,isactive,argc,argv)
{
#ifdef __MINGW32__
  return __FAIL;
#else
  long s;
  if (argc == 1 && isint(argv[0], &s))
    return mkfalse;
  else if (argc == 1 && isvoid(argv[0]))
    return mktrue;
  else
    return __FAIL;
#endif
}

FUNCTION(clib,isexited,argc,argv)
{
#ifdef __MINGW32__
  return __FAIL;
#else
  long s;
  if (argc == 1 && isint(argv[0], &s)) {
    int status = s;
    if (WIFEXITED(status))
      return mktrue;
    else
      return mkfalse;
  } else if (argc == 1 && isvoid(argv[0]))
    return mkfalse;
  else
    return __FAIL;
#endif
}

FUNCTION(clib,exitstatus,argc,argv)
{
#ifdef __MINGW32__
  return __FAIL;
#else
  long s;
  if (argc == 1 && isint(argv[0], &s)) {
    int status = s;
    if (WIFEXITED(status))
      return mkint(WEXITSTATUS(status));
    else
      return __FAIL;
  } else
    return __FAIL;
#endif
}

FUNCTION(clib,issignaled,argc,argv)
{
#ifdef __MINGW32__
  return __FAIL;
#else
  long s;
  if (argc == 1 && isint(argv[0], &s)) {
    int status = s;
    if (WIFSIGNALED(status))
      return mktrue;
    else
      return mkfalse;
  } else if (argc == 1 && isvoid(argv[0]))
    return mkfalse;
  else
    return __FAIL;
#endif
}

FUNCTION(clib,termsig,argc,argv)
{
#ifdef __MINGW32__
  return __FAIL;
#else
  long s;
  if (argc == 1 && isint(argv[0], &s)) {
    int status = s;
    if (WIFSIGNALED(status))
      return mkint(WTERMSIG(status));
    else
      return __FAIL;
  } else
    return __FAIL;
#endif
}

FUNCTION(clib,isstopped,argc,argv)
{
#ifdef __MINGW32__
  return __FAIL;
#else
  long s;
  if (argc == 1 && isint(argv[0], &s)) {
    int status = s;
    if (WIFSTOPPED(status))
      return mktrue;
    else
      return mkfalse;
  } else if (argc == 1 && isvoid(argv[0]))
    return mkfalse;
  else
    return __FAIL;
#endif
}

FUNCTION(clib,stopsig,argc,argv)
{
#ifdef __MINGW32__
  return __FAIL;
#else
  long s;
  if (argc == 1 && isint(argv[0], &s)) {
    int status = s;
    if (WIFSTOPPED(status))
      return mkint(WSTOPSIG(status));
    else
      return __FAIL;
  } else
    return __FAIL;
#endif
}

FUNCTION(clib,getenv,argc,argv)
{
  char *name, *val;
  if (argc == 1 && isstr(argv[0], &name))
    if ((val = getenv(name))) {
      char *s = strdup(val);
      if (s)
	return mkstr(s);
      else
	return __ERROR;
    } else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(clib,setenv,argc,argv)
{
  char *name, *val;
  if (argc == 2 && isstr(argv[0], &name) && isstr(argv[1], &val)) {
    /* According to POSIX, the string passed to putenv itself becomes
       part of the environment, so it has to be allocated dynamically. */
    char *envstr = malloc(strlen(name)+strlen(val)+2);
    if (envstr)
      sprintf(envstr, "%s=%s", name, val);
    else
      return __ERROR;
    if (putenv(envstr))
      return __FAIL;
    else
      return mkvoid;
  } else
    return __FAIL;
}

/* multithreading: ********************************************************/

/* the thread table (make sure that the size matches with qdefs.h) */

#ifdef USE_THREADS
#define MAXTHREAD 1024
#else
#define MAXTHREAD 1
#endif

typedef unsigned char bool;
typedef struct {
  bool			active, canceled, used;	/* status */
  expr			arg;			/* thread argument */
  expr			result;			/* thread result, if any */
  expr			thread;			/* thread object */
#ifdef USE_THREADS
  pthread_t		id;			/* thread id */
  pthread_mutex_t	exit_mutex;		/* exit mutex and condition */
  pthread_cond_t	exit_cond;
#endif
} THREAD;

static THREAD threads[MAXTHREAD], *thr0 = threads;
static int maxused, stamp;

DESTRUCTOR(clib,Thread,ptr)
{
#ifdef USE_THREADS
  THREAD *thr = (THREAD*)ptr;
  pthread_t id = thr->id;
  if (thr == thr0) {
    if (thr->thread) thr->thread = NULL;
    return;
  }
#ifdef DEBUG
  printf("destroying thread %d (id %d)\n", thr-thr0, (int)id);
#endif
  pthread_cancel(id);
  /* wait for the canceled thread to finish */
  release_lock();
  pthread_join(id, NULL);
  acquire_lock();
  thr->active = thr->canceled = thr->used = 0;
  if (thr->arg) {
    freeref(thr->arg);
    thr->arg = NULL;
  }
  if (thr->result) {
    freeref(thr->result);
    thr->result = NULL;
  }
  if (thr->thread) thr->thread = NULL;
  pthread_mutex_destroy(&thr->exit_mutex);
  pthread_cond_destroy(&thr->exit_cond);
  /* deregister the thread with the interpreter */
  fini_thread(thr-thr0);
#endif
}

FUNCTION(clib,thread_no,argc,argv)
{
  THREAD *thr;
  if (argc == 1 && isobj(argv[0], type(Thread), (void**)&thr)) {
    return mkint(thr-thr0);
  } else
    return __FAIL;
}

FUNCTION(clib,this_thread,argc,argv)
{
  if (argc == 0) {
    THREAD *thr = thr0+this_thread();
    if (thr->thread)
      return thr->thread;
    else if (thr == thr0)
      return (thr->thread = mkobj(type(Thread), thr));
    else
      return __FAIL;
  } else
    return __FAIL;
}

#ifdef USE_THREADS

static bool thread_ready;
static pthread_mutex_t thread_ready_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t thread_ready_cond = PTHREAD_COND_INITIALIZER;
static THREAD *nthr;

static void my_mutex_unlock(void *mut)
{
  pthread_mutex_unlock((pthread_mutex_t*)mut);
}

static void thread_canceled_proc(void *arg)
{
  THREAD *thr = (THREAD*)arg;
#ifdef DEBUG
  printf("thread %d (id %d) %s\n", thr-thr0, (int)thr->id,
	 thr->result?"exited":"canceled");
#endif
  thr->canceled = (thr->result == NULL);
  exit_thread(thr-thr0);
  pthread_mutex_lock(&thr->exit_mutex);
  thr->active = 0;
  pthread_cond_broadcast(&thr->exit_cond);
  pthread_mutex_unlock(&thr->exit_mutex);
}

static void *thread_proc(void *arg)
{
  expr x = (expr)arg, y;
  THREAD *thr;
  /* register the new thread with the interpreter */
  int thrid = init_thread();
  /* fill in the thread info */
  pthread_mutex_lock(&thread_ready_mutex);
  if (thrid >= 0) {
    thr = thr0+thrid;
    if ((thr->thread = mkobj(type(Thread), thr))) {
      thr->arg = x;
      thr->result = NULL;
      thr->id = pthread_self();
      pthread_mutex_init(&thr->exit_mutex, NULL);
      pthread_cond_init(&thr->exit_cond, NULL);
      thr->active = 1; thr->canceled = 0; thr->used = 1;
      if (thr-thr0 > maxused) maxused = thr-thr0;
      nthr = thr;
#ifdef DEBUG
      printf("initialized thread %d (id %d)\n", thr-thr0, (int)thr->id);
#endif
    } else {
      exit_thread(thrid);
      fini_thread(thrid);
#ifdef DEBUG
      printf("error initializing thread %d (id %d), exiting\n", thr-thr0, (int)thr->id);
#endif
    }
  }
#ifdef DEBUG
  else
    printf("error registering thread\n");
#endif
  /* exit if error */
  if (!nthr) {
    thread_ready = 1;
    pthread_cond_signal(&thread_ready_cond);
    pthread_mutex_unlock(&thread_ready_mutex);
    return NULL;
  }
  /* signal that we're ready */
  pthread_cleanup_push(thread_canceled_proc, thr);
  release_lock();
  thread_ready = 1;
  pthread_cond_signal(&thread_ready_cond);
  pthread_mutex_unlock(&thread_ready_mutex);
  /* we're up and running, start evaluating the special argument */
#ifdef DEBUG
  printf("thread %d (id %d) up and running\n", thr-thr0, (int)thr->id);
#endif
  acquire_lock();
  y = newref(eval(x));
  pthread_cleanup_pop(0);
  /* exit from thread */
  thr->result = y;
  exit_thread(thr-thr0);
  /* signal that we've exited */
  pthread_mutex_lock(&thr->exit_mutex);
  thr->active = 0;
  pthread_cond_broadcast(&thr->exit_cond);
  pthread_mutex_unlock(&thr->exit_mutex);
#ifdef DEBUG
  printf("thread %d (id %d) exits\n", thr-thr0, (int)thr->id);
#endif
  return (void*)y;
}
#endif

FUNCTION(clib,thread,argc,argv)
{
#ifdef USE_THREADS
  if (argc == 1) {
    pthread_t id;
    int res;
    /* start the new thread */
    pthread_cleanup_push(my_mutex_unlock, (void*)&thread_ready_mutex);
    pthread_mutex_lock(&thread_ready_mutex);
    thread_ready = 0; nthr = NULL;
#ifdef DEBUG
    printf("starting new thread\n");
#endif
    if ((res = pthread_create(&id, NULL, thread_proc, newref(argv[0]))) == 0) {
      /* wait until the new thread signals that it's up and running */
#ifdef DEBUG
      printf("waiting for new thread\n");
#endif
      release_lock();
      while (!thread_ready)
	pthread_cond_wait(&thread_ready_cond, &thread_ready_mutex);
      acquire_lock();
    }
    pthread_cleanup_pop(1);
    if (res) return __FAIL;
#ifdef DEBUG
    printf("new thread %s\n", nthr?"ready":"aborted");
#endif
    /* return the thread object */
    if (nthr)
      return nthr->thread;
    else
      return __ERROR;
  } else
#endif
    return __FAIL;
}

FUNCTION(clib,return,argc,argv)
{
  if (argc == 1) {
    THREAD *thr = thr0+this_thread();
#ifdef USE_THREADS
    if (thr > thr0) {
      thr->result = newref(argv[0]);
      pthread_exit((void*)thr->result);
    } else
#endif
      return mksym(sym(halt));
  } else
    return __FAIL;
}

FUNCTION(clib,cancel,argc,argv)
{
#ifdef USE_THREADS
  THREAD *thr;
  if (argc == 1 && isobj(argv[0], type(Thread), (void**)&thr)) {
    if (thr == thr0) return __FAIL;
    pthread_cancel(thr->id);
    return mkvoid;
  } else
#endif
    return __FAIL;
}

FUNCTION(clib,result,argc,argv)
{
#ifdef USE_THREADS
  THREAD *thr;
  if (argc == 1 && isobj(argv[0], type(Thread), (void**)&thr)) {
    if (thr == thr0) return __FAIL;
    /* wait for the thread to finish */
    pthread_cleanup_push(my_mutex_unlock, (void*)&thr->exit_mutex);
    pthread_mutex_lock(&thr->exit_mutex);
    release_lock();
    while (thr->active)
      pthread_cond_wait(&thr->exit_cond, &thr->exit_mutex);
    pthread_cleanup_pop(1);
    acquire_lock();
    /* return result, if any */
    if (thr->canceled || !thr->result)
      return __FAIL;
    else
      return thr->result;
  } else
#endif
    return __FAIL;
}

FUNCTION(clib,yield,argc,argv)
{
  if (argc == 0) {
#ifdef USE_THREADS
    release_lock(); acquire_lock();
#endif
    return mkvoid;
  } else
    return __FAIL;
}

FUNCTION(clib,active,argc,argv)
{
  THREAD *thr;
  if (argc == 1 && isobj(argv[0], type(Thread), (void**)&thr))
#ifdef USE_THREADS
    return thr->active?mktrue:mkfalse;
#else
    return mktrue;
#endif
  else
    return __FAIL;
}

FUNCTION(clib,canceled,argc,argv)
{
  THREAD *thr;
  if (argc == 1 && isobj(argv[0], type(Thread), (void**)&thr))
#ifdef USE_THREADS
    return thr->canceled?mktrue:mkfalse;
#else
    return mkfalse;
#endif
  else
    return __FAIL;
}

FUNCTION(clib,schedule,argc,argv)
{
#ifdef USE_THREADS
  THREAD *thr;
  long pol, prio;
  if (argc == 3 && isobj(argv[0], type(Thread), (void**)&thr) &&
      isint(argv[1], &pol) && isint(argv[1], &prio)) {
    struct sched_param param;
    int actpol;
    switch (pol) {
    case 0: pol = SCHED_OTHER; prio = 0; break;
    case 1: pol = SCHED_RR; break;
    case 2: pol = SCHED_FIFO; break;
    default: return __FAIL;
    }
    if (pthread_getschedparam(thr->id, &actpol, &param))
      return __FAIL;
    else {
      param.sched_priority = prio;
      if (pthread_setschedparam(thr->id, pol, &param))
	return __FAIL;
      else
	return mkvoid;
    }
  } else
#endif
    return __FAIL;
}

#ifdef USE_THREADS

/* mutex data structure */

typedef struct {
  pthread_mutex_t mut;
  pthread_mutexattr_t attr, *attrp;
  int stamp;
} my_mutex_t;

/* condition data structure */

typedef struct {
  pthread_mutex_t mut;
  pthread_cond_t cond;
  bool set;
  int stamp;
} my_cond_t;

/* semaphore data structure */

typedef struct expr_queue_entry {
  expr val;
  struct expr_queue_entry *next;
} expr_queue_entry_t;

typedef struct {
  long size;
  expr_queue_entry_t *head, *tail, *last_tail;
} expr_queue_t;

typedef struct {
  pthread_mutex_t mut;
  sem_t sem;
  pthread_cond_t cond;
  expr_queue_t queue;
  int stamp;
  long max;
} my_sem_t;

static void init_queue(expr_queue_t *queue)
{
  queue->size = 0;
  queue->head = queue->tail = queue->last_tail = NULL;
}

static void free_queue(expr_queue_t *queue)
{
  expr_queue_entry_t *head = queue->head;
  while (head) {
    expr_queue_entry_t *next = head->next;
    if (head->val) freeref(head->val);
    free(head);
    head = next;
  }
}

static expr_queue_entry_t *enqueue_expr(expr_queue_t *queue, expr val)
{
  expr_queue_entry_t *new_tail = malloc(sizeof(expr_queue_entry_t));
  if (!new_tail) return NULL;
  new_tail->val = newref(val);
  new_tail->next = NULL;
  if (queue->tail) {
    queue->tail->next = new_tail;
    queue->last_tail = queue->tail;
    queue->tail = new_tail;
  } else {
    queue->last_tail = NULL;
    queue->head = queue->tail = new_tail;
  }
  queue->size++;
  return new_tail;
}

static expr_queue_entry_t *unenqueue_expr(expr_queue_t *queue)
{
  if (queue->tail) {
    freeref(queue->tail->val);
    free(queue->tail);
    queue->tail = queue->last_tail;
    if (queue->tail)
      queue->tail->next = NULL;
    else
      queue->head = NULL;
    queue->last_tail = NULL;
    queue->size--;
    return queue->tail;
  } else
    return NULL;
}

static expr dequeue_expr(expr_queue_t *queue)
{
  expr val;
  if (!queue->head) return NULL;
  val = queue->head->val;
  if (queue->head == queue->tail) {
    free(queue->head);
    queue->head = queue->tail = queue->last_tail = NULL;
  } else {
    expr_queue_entry_t *next = queue->head->next;
    if (queue->last_tail == queue->head) queue->last_tail = NULL;
    free(queue->head);
    queue->head = next;
  }
  queue->size--;
  return val;
}

/* clean up after fork */

static check_mut(my_mutex_t *mut)
{
  if (mut && mut->stamp != stamp) {
    pthread_mutex_init(&mut->mut, mut->attrp);
    mut->stamp = stamp;
  }
}

static check_cond(my_cond_t *cond)
{
  if (cond && cond->stamp != stamp) {
    pthread_mutex_init(&cond->mut, NULL);
    pthread_cond_init(&cond->cond, NULL);
    cond->set = 0;
    cond->stamp = stamp;
  }
}

static check_sem(my_sem_t *sem)
{
  if (sem && sem->stamp != stamp) {
    pthread_mutex_init(&sem->mut, NULL);
    sem_init(&sem->sem, 0, sem->queue.size);
    pthread_cond_init(&sem->cond, NULL);
    sem->stamp = stamp;
  }
}

#endif

/* destructors */

DESTRUCTOR(clib,Mutex,ptr)
{
#ifdef USE_THREADS
  my_mutex_t *mut = (my_mutex_t*)ptr;
  if (mut) {
    pthread_mutex_destroy(&mut->mut);
    if (mut->attrp)
      pthread_mutexattr_destroy(mut->attrp);
    free(mut);
  }
#endif
}

DESTRUCTOR(clib,Cond,ptr)
{
#ifdef USE_THREADS
  my_cond_t *cond = (my_cond_t*)ptr;
  if (cond) {
    pthread_mutex_destroy(&cond->mut);
    pthread_cond_destroy(&cond->cond);
    free(cond);
  }
#endif
}

DESTRUCTOR(clib,Sem,ptr)
{
#ifdef USE_THREADS
  my_sem_t *sem = (my_sem_t*)ptr;
  if (sem) {
    pthread_mutex_destroy(&sem->mut);
    sem_destroy(&sem->sem);
    pthread_cond_destroy(&sem->cond);
    free_queue(&sem->queue);
    free(sem);
  }
#endif
}

/* constructors */

FUNCTION(clib,mutex,argc,argv)
{
#ifdef USE_THREADS
  my_mutex_t *mut = (my_mutex_t*)malloc(sizeof(my_mutex_t));
  if (mut) {
    mut->attrp = NULL;
    pthread_mutex_init(&mut->mut, NULL);
    mut->stamp = stamp;
    return mkobj(type(Mutex), mut);
  } else
    return __ERROR;
#else
  return __FAIL;
#endif
}

FUNCTION(clib,errorchecking_mutex,argc,argv)
{
#ifdef USE_THREADS
  my_mutex_t *mut = (my_mutex_t*)malloc(sizeof(my_mutex_t));
  if (mut) {
    mut->attrp = &mut->attr;
    pthread_mutexattr_init(mut->attrp);
    pthread_mutexattr_settype(mut->attrp, PTHREAD_MUTEX_ERRORCHECK);
    pthread_mutex_init(&mut->mut, mut->attrp);
    mut->stamp = stamp;
    return mkobj(type(Mutex), mut);
  } else
    return __ERROR;
#else
  return __FAIL;
#endif
}

FUNCTION(clib,recursive_mutex,argc,argv)
{
#ifdef USE_THREADS
  my_mutex_t *mut = (my_mutex_t*)malloc(sizeof(my_mutex_t));
  if (mut) {
    mut->attrp = &mut->attr;
    pthread_mutexattr_init(mut->attrp);
    pthread_mutexattr_settype(mut->attrp, PTHREAD_MUTEX_RECURSIVE);
    pthread_mutex_init(&mut->mut, mut->attrp);
    mut->stamp = stamp;
    return mkobj(type(Mutex), mut);
  } else
    return __ERROR;
#else
  return __FAIL;
#endif
}

FUNCTION(clib,cond,argc,argv)
{
#ifdef USE_THREADS
  my_cond_t *cond = (my_cond_t*)malloc(sizeof(my_cond_t));
  if (cond) {
    pthread_mutex_init(&cond->mut, NULL);
    pthread_cond_init(&cond->cond, NULL);
    cond->set = 0;
    cond->stamp = stamp;
    return mkobj(type(Cond), cond);
  } else
    return __ERROR;
#else
  return __FAIL;
#endif
}

FUNCTION(clib,sem,argc,argv)
{
#ifdef USE_THREADS
  my_sem_t *sem = (my_sem_t*)malloc(sizeof(my_sem_t));
  if (sem) {
    pthread_mutex_init(&sem->mut, NULL);
    sem_init(&sem->sem, 0, 0);
    pthread_cond_init(&sem->cond, NULL);
    init_queue(&sem->queue);
    sem->stamp = stamp;
    sem->max = 0;
    return mkobj(type(Sem), sem);
  } else
    return __ERROR;
#else
  return __FAIL;
#endif
}

FUNCTION(clib,bounded_sem,argc,argv)
{
#ifdef USE_THREADS
  long max;
  if (argc == 1 && isint(argv[0], &max) && max > 0) {
    my_sem_t *sem = (my_sem_t*)malloc(sizeof(my_sem_t));
    if (sem) {
      pthread_mutex_init(&sem->mut, NULL);
      sem_init(&sem->sem, 0, 0);
      pthread_cond_init(&sem->cond, NULL);
      init_queue(&sem->queue);
      sem->stamp = stamp;
      sem->max = max;
      return mkobj(type(Sem), sem);
    } else
      return __ERROR;
  } else
#endif
    return __FAIL;
}

/* interface ops */

FUNCTION(clib,lock,argc,argv)
{
#ifdef USE_THREADS
  my_mutex_t *mut;
  if (argc == 1 && isobj(argv[0], type(Mutex), (void**)&mut)) {
    int res;
    check_mut(mut);
    release_lock();
    res = pthread_mutex_lock(&mut->mut);
    acquire_lock();
    if (res)
      return __FAIL;
    else
      return mkvoid;
  } else
#endif
    return __FAIL;
}

FUNCTION(clib,unlock,argc,argv)
{
#ifdef USE_THREADS
  my_mutex_t *mut;
  if (argc == 1 && isobj(argv[0], type(Mutex), (void**)&mut)) {
    check_mut(mut);
    if (pthread_mutex_unlock(&mut->mut))
      return __FAIL;
    else
      return mkvoid;
  } else
#endif
    return __FAIL;
}

FUNCTION(clib,try,argc,argv)
{
#ifdef USE_THREADS
  my_mutex_t *mut;
  my_sem_t *sem;
  if (argc == 1)
    if (isobj(argv[0], type(Mutex), (void**)&mut)) {
      check_mut(mut);
      if (pthread_mutex_trylock(&mut->mut))
	return __FAIL;
      else
	return mkvoid;
    } else if (isobj(argv[0], type(Sem), (void**)&sem)) {
      int res;
      expr val;
      check_sem(sem);
      release_lock();
      res = sem_trywait(&sem->sem);
      if (!res) {
	pthread_mutex_lock(&sem->mut);
	if (sem->queue.size <= 0)
	  res = -1;
	else {
	  val = dequeue_expr(&sem->queue);
	  if (sem->max) pthread_cond_signal(&sem->cond);
	}
	pthread_mutex_unlock(&sem->mut);
      }
      acquire_lock();
      if (res)
	return __FAIL;
      else
	return unref(val);
    } else
      return __FAIL;
  else
#endif
    return __FAIL;
}

FUNCTION(clib,signal,argc,argv)
{
#ifdef USE_THREADS
  my_cond_t *cond;
  if (argc == 1 && isobj(argv[0], type(Cond), (void**)&cond)) {
    int res;
    check_cond(cond);
    pthread_mutex_lock(&cond->mut);
    res = pthread_cond_signal(&cond->cond);
    if (!res) cond->set = 1;
    pthread_mutex_unlock(&cond->mut);
    if (res)
      return __FAIL;
    else
      return mkvoid;
  } else
#endif
    return __FAIL;
}

FUNCTION(clib,broadcast,argc,argv)
{
#ifdef USE_THREADS
  my_cond_t *cond;
  if (argc == 1 && isobj(argv[0], type(Cond), (void**)&cond)) {
    int res;
    check_cond(cond);
    pthread_mutex_lock(&cond->mut);
    res = pthread_cond_broadcast(&cond->cond);
    if (!res) cond->set = 1;
    pthread_mutex_unlock(&cond->mut);
    if (res)
      return __FAIL;
    else
      return mkvoid;
  } else
#endif
    return __FAIL;
}

FUNCTION(clib,suspend,argc,argv)
{
#ifdef USE_THREADS
  my_cond_t *cond;
  int n;
  expr *v;
  double t;
  struct timespec tspec;
  bool tset = 0;

  if (argc == 1 &&
      (isobj(argv[0], type(Cond), (void**)&cond) ||
       istuple(argv[0], &n, &v) && n == 2 && (tset = 1) &&
       isobj(v[0], type(Cond), (void**)&cond) &&
       (isfloat(v[1], &t) || ismpz_float(v[1], &t)))) {
    int res = 0;
    check_cond(cond);
    if (tset) {
      double ip, fp;
      unsigned long secs;
      unsigned long nsecs;
      fp = modf(t/1e3, &ip);
      if (ip > LONG_MAX) { ip = (double)LONG_MAX; fp = 0.0; }
      secs = (unsigned long)ip;
      nsecs = (unsigned long)(fp*1e9);
      tspec.tv_sec = secs; tspec.tv_nsec = nsecs;
    }
    pthread_cleanup_push(my_mutex_unlock, (void*)&cond->mut);
    pthread_mutex_lock(&cond->mut);
    release_lock();
    cond->set = 0;
    while (!cond->set && !res)
      if (tset)
	res = pthread_cond_timedwait(&cond->cond, &cond->mut, &tspec);
      else
	res = pthread_cond_wait(&cond->cond, &cond->mut);
    pthread_cleanup_pop(1);
    acquire_lock();
    if (res)
      return __FAIL;
    else
      return mkvoid;
  } else
#endif
    return __FAIL;
}

FUNCTION(clib,post,argc,argv)
{
#ifdef USE_THREADS
  my_sem_t *sem;
  if (argc == 2 && isobj(argv[0], type(Sem), (void**)&sem)) {
    int res = 0;
    check_sem(sem);
    if (sem->max) {
      pthread_cleanup_push(my_mutex_unlock, (void*)&sem->mut);
      pthread_mutex_lock(&sem->mut);
      release_lock();
      while (sem->queue.size >= sem->max && !res)
	res = pthread_cond_wait(&sem->cond, &sem->mut);
      if (!res) {
	if (!enqueue_expr(&sem->queue, argv[1]))
	  res = -1;
	if (!res) {
	  res = sem_post(&sem->sem);
	  if (res) unenqueue_expr(&sem->queue);
	}
      }
      pthread_cleanup_pop(1);
      acquire_lock();
    } else {
      pthread_mutex_lock(&sem->mut);
      if (!enqueue_expr(&sem->queue, argv[1]))
	res = -1;
      if (!res) {
	res = sem_post(&sem->sem);
	if (res) unenqueue_expr(&sem->queue);
      }
      pthread_mutex_unlock(&sem->mut);
    }
    if (res)
      return (res==-1)?__ERROR:__FAIL;
    else
      return mkvoid;
  } else
#endif
    return __FAIL;
}

FUNCTION(clib,get,argc,argv)
{
#ifdef USE_THREADS
  my_sem_t *sem;
  if (argc == 1 && isobj(argv[0], type(Sem), (void**)&sem)) {
    int res;
    expr val;
    check_sem(sem);
    release_lock();
  retry:
    res = sem_wait(&sem->sem);
    if (!res) {
      pthread_mutex_lock(&sem->mut);
      if (sem->queue.size <= 0) {
	pthread_mutex_unlock(&sem->mut);
	goto retry;
      }
      val = dequeue_expr(&sem->queue);
      if (sem->max) pthread_cond_signal(&sem->cond);
      pthread_mutex_unlock(&sem->mut);
    }
    acquire_lock();
    if (res)
      return __FAIL;
    else
      return unref(val);
  } else
#endif
    return __FAIL;
}

FUNCTION(clib,get_size,argc,argv)
{
#ifdef USE_THREADS
  my_sem_t *sem;
  if (argc == 1 && isobj(argv[0], type(Sem), (void**)&sem)) {
    long size;
    check_sem(sem);
    pthread_mutex_lock(&sem->mut);
    size = sem->queue.size;
    pthread_mutex_unlock(&sem->mut);
    return mkint(size);
  } else
#endif
    return __FAIL;
}

FUNCTION(clib,get_bound,argc,argv)
{
#ifdef USE_THREADS
  my_sem_t *sem;
  if (argc == 1 && isobj(argv[0], type(Sem), (void**)&sem)) {
    check_sem(sem);
    return mkint(sem->max);
  } else
#endif
    return __FAIL;
}

/* time functions: ********************************************************/

static int tzset_init = 0;

FUNCTION(clib,timezone,argc,argv)
{
  if (argc == 0) {
    tzset(); tzset_init = 1;
    return mkstr(strdup(tzname[0]));
  } else
    return __FAIL;
}

FUNCTION(clib,ctime,argc,argv)
{
  if (argc == 1) {
    long it, l;
    double ft;
    time_t t;
    char *st;
    if (isfloat(argv[0], &ft))
      if (ft < (double)INT_MIN || ft > (double)INT_MAX)
	return __FAIL;
      else
	it = (int)ft;
    else if (!isint(argv[0], &it))
      return __FAIL;
    t = it;
    /* some systems forget to call tzset() in localtime(), take care of that */
    if (!tzset_init) { tzset(); tzset_init = 1; }
    st = asctime(localtime(&t));
    l = strlen(st);
    if (l > 0 && st[l-1] == '\n') st[l-1] = 0;
    return mkstr(strdup(st));
  } else
    return __FAIL;
}

FUNCTION(clib,gmtime,argc,argv)
{
  if (argc == 1) {
    long it, l;
    double ft;
    time_t t;
    char *st;
    if (isfloat(argv[0], &ft))
      if (ft < (double)INT_MIN || ft > (double)INT_MAX)
	return __FAIL;
      else
	it = (int)ft;
    else if (!isint(argv[0], &it))
      return __FAIL;
    t = it;
    st = asctime(gmtime(&t));
    l = strlen(st);
    if (l > 0 && st[l-1] == '\n') st[l-1] = 0;
    return mkstr(strdup(st));
  } else
    return __FAIL;
}

FUNCTION(clib,clock,argc,argv)
{
  if (argc == 0) {
    clock_t t = clock();
    if (t == (clock_t)-1)
      return __FAIL;
    else
      return mkint((long)t);
  } else
    return __FAIL;
}

FUNCTION(clib,clocks_per_sec,argc,argv)
{
  if (argc == 0)
    return mkint((long)CLOCKS_PER_SEC);
  else
    return __FAIL;
}

/* filename globbing: *****************************************************/

FUNCTION(clib,fnmatch,argc,argv)
{
  char *pattern, *s;
  if (argc == 2 && isstr(argv[0], &pattern) && isstr(argv[1], &s))
    if (fnmatch(pattern, s, 0))
      return mkfalse;
    else
      return mktrue;
  else
    return __FAIL;
}

FUNCTION(clib,glob,argc,argv)
{
  char *pattern;
  if (argc == 1 && isstr(argv[0], &pattern)) {
    glob_t g;
    int res = glob(pattern, 0, NULL, &g);
    if (res == GLOB_NOMATCH)
      return mknil;
    else if (res)
      return __FAIL;
    else {
      expr x = mknil;
      int i = g.gl_pathc;
      while (x && --i >= 0)
	x = mkcons(mkstr(strdup(g.gl_pathv[i])), x);
      globfree(&g);
      if (x)
	return x;
      else
	return __ERROR;
    }
  } else
    return __FAIL;
}

/* regular expression matching: *******************************************/

/* regexp stack */

typedef struct {
  unsigned done:1, global:2, matched:1;
  int cflags, eflags;
  regex_t rx;
  regmatch_t *matches;
  char *s, *p, *start;
} regstate_t;

long regalloc = 0;
regstate_t *regstack = NULL, *regp = NULL;
char regmsg[BUFSZ];

#define REGALLOC 50

static int reg_push(void)
{
  if (!regstack)
    if ((regstack = malloc(REGALLOC*sizeof(regstate_t)))) {
      regalloc = REGALLOC;
      regp = regstack;
    } else
      return -1;
  else if (!regp)
    regp = regstack;
  else if (regp-regstack+1 == regalloc) {
    regstate_t *newstack = realloc(regstack,
				   (regalloc+REGALLOC)*sizeof(regstate_t));
    if (newstack) {
      regstack = newstack;
      regp = regstack+regalloc;
      regalloc += REGALLOC;
    } else
      return -1;
  } else
    regp++;
  regp->done = regp->global = regp->matched = 0;
  regp->cflags = regp->eflags = 0;
  regp->matches = NULL;
  regp->s = regp->p = regp->start = NULL;
  return 0;
}

static void reg_pop(void)
{
  if (!regp) return;
  regfree(&regp->rx);
  if (regp->matches) free(regp->matches);
  if (regp->s) free(regp->s);
  if (regp > regstack)
    regp--;
  else
    regp = NULL;
}

/* push a new expression on the stack */

static int reg_add(char *pattern, char *s, int global, int cflags, int eflags)
{
  int ret;
  if (regp && regp->done) reg_pop();
  if (reg_push()) return -1;
  regp->global = global;
  regp->cflags = cflags;
  regp->eflags = eflags;
  ret = regcomp(&regp->rx, pattern, REG_EXTENDED|cflags);
  *regmsg = 0;
  if (ret) {
    regerror(ret, &regp->rx, regmsg, BUFSZ);
    reg_pop();
    return ret;
  }
  if (!(regp->s = strdup(s))) {
    reg_pop();
    return -1;
  }
  regp->p = regp->s; regp->start = NULL;
  if (!(regp->matches = malloc((regp->rx.re_nsub+1)*sizeof(regmatch_t)))) {
    reg_pop();
    return -1;
  }
  return 0;
}

/* search */

static int reg_flags(char *p)
{
  int flags;
  flags = regp->eflags;
  if (p > regp->s)
    if (regp->cflags & REG_NEWLINE)
      if (p[-1] == '\n')
	flags &= ~REG_NOTBOL;
      else
	flags |= REG_NOTBOL;
    else
      flags |= REG_NOTBOL;
  return flags;
}

static int reg_search(void)
{
  int ret;
  char *prev;
  while (regp && regp->done && regp>regstack) reg_pop();
  if (!regp) return -1;
  if (regp->matched)
    /* note the beginning of the previous match */
    prev = regp->start+regp->matches[0].rm_so;
  regp->start = regp->p;
  if (regp->global || !regp->matched) {
    ret = regexec(&regp->rx, regp->p, regp->rx.re_nsub+1, regp->matches,
		  reg_flags(regp->p));
    if (!ret) {
      if (regp->matched)
	if (regp->matches[0].rm_eo == regp->matches[0].rm_so &&
	    regp->p == prev)
	  /* an extra empty match: if not at end of string then advance to the
	     next position and try again; otherwise simply ignore this match
	     and fail */
	  if (*regp->p) {
	    int i;
	    /* this cannot fail since we can always match the empty string */
	    ret = regexec(&regp->rx, regp->p+1, regp->rx.re_nsub+1,
			  regp->matches, reg_flags(regp->p+1));
	    /* translate offsets */
	    for (i = 0; i <= regp->rx.re_nsub; i++) {
	      regp->matches[i].rm_so++;
	      regp->matches[i].rm_eo++;
	    }
	  } else
	    ret = REG_NOMATCH;
      regp->matched = 1;
    }
  } else
    ret = REG_NOMATCH;
  *regmsg = 0;
  if (ret) {
    regp->done = 1;
    regerror(ret, &regp->rx, regmsg, BUFSZ);
  } else if (regp->global == 2 &&
	     regp->matches[0].rm_eo > regp->matches[0].rm_so)
    regp->p += regp->matches[0].rm_so+1;
  else
    regp->p += regp->matches[0].rm_eo;
  return ret;
}

/* stop search */

static void reg_done(void)
{
  if (regp) {
    regp->start = regp->p;
    regp->done = 1;
  }
}

/* return matches */

static size_t reg_nmatches(void)
{
  if (regp)
    return regp->rx.re_nsub;
  else
    return 0;
}

static long reg_start(void)
{
  if (regp && regp->start)
    return regp->start-regp->s;
  else
    return -1;
}

static char *reg_skipstr(void)
{
  if (regp && regp->start)
    return regp->start;
  else
    return NULL;
}

static long reg_pos(int i)
{
  if (regp && regp->start && 0 <= i && i <= regp->rx.re_nsub)
    if (!regp->done && regp->matches[i].rm_so >= 0)
      return regp->start+regp->matches[i].rm_so-regp->s;
    else
      return -1;
  else
    return -1;
}

static long reg_end(int i)
{
  if (regp && regp->start && 0 <= i && i <= regp->rx.re_nsub)
    if (!regp->done && regp->matches[i].rm_eo >= 0)
      return regp->start+regp->matches[i].rm_eo-regp->s;
    else
      return -1;
  else
    return -1;
}

static char *reg_str(int i)
{
  if (regp && regp->start && 0 <= i && i <= regp->rx.re_nsub)
    if (!regp->done && regp->matches[i].rm_so >= 0)
      return regp->start+regp->matches[i].rm_so;
    else
      return NULL;
  else
    return NULL;
}

/* interface functions */

FUNCTION(clib,regmatch,argc,argv)
{
  char *opts, *regex, *s;
  int cflags = 0, eflags = 0, global = 0, ret;
  if (argc != 3 || !isstr(argv[0], &opts) || !isstr(argv[1], &regex) ||
      !isstr(argv[2], &s))
    return __FAIL;
  while (*opts)
    switch (*(opts++)) {
    case 'g':
      if (!global) global = 1;
      break;
    case 'G':
      global = 2;
      break;
    case 'i':
      cflags |= REG_ICASE;
      break;
    case 'n':
      cflags |= REG_NEWLINE;
      break;
    case '^':
      eflags |= REG_NOTBOL;
      break;
    case '$':
      eflags |= REG_NOTEOL;
      break;
    default:
      return __FAIL;
    }
  ret = reg_add(regex, s, global, cflags, eflags);
  if (ret == -1)
    return __ERROR;
  else if (ret)
    return mkapp(mksym(sym(regerr)), mkstr(strdup(regmsg)));
  ret = reg_search();
  if (ret == -1 || ret == REG_NOMATCH)
    return mkfalse;
  else if (ret)
    return mkapp(mksym(sym(regerr)), mkstr(strdup(regmsg)));
  else
    return mktrue;
}

FUNCTION(clib,regnext,argc,argv)
{
  int ret;
  if (argc != 0) return __FAIL;
  ret = reg_search();
  if (ret == -1 || ret == REG_NOMATCH)
    return mkfalse;
  else if (ret)
    return mkapp(mksym(sym(regerr)), mkstr(strdup(regmsg)));
  else
    return mktrue;
}

FUNCTION(clib,regdone,argc,argv)
{
  if (argc != 0) return __FAIL;
  reg_done();
  return mkvoid;
}

FUNCTION(clib,regstart,argc,argv)
{
  long start;
  if (argc != 0) return __FAIL;
  start = reg_start();
  if (start >= 0)
    return mkint(start);
  else
    return __FAIL;
}

FUNCTION(clib,regskip,argc,argv)
{
  char *skip;
  if (argc != 0) return __FAIL;
  if ((skip = reg_skipstr())) {
    long start = reg_start(), pos = reg_pos(0);
    char *s;
    if (pos >= start)
      s = malloc(pos-start+1);
    else
      s = malloc(strlen(skip)+1);
    if (!s) return __ERROR;
    if (pos >= start) {
      strncpy(s, skip, pos-start);
      s[pos-start] = 0;
    } else
      strcpy(s, skip);
    return mkstr(s);
  } else
    return __FAIL;
}

FUNCTION(clib,reg,argc,argv)
{
  long i;
  if (argc != 1 || !isint(argv[0], &i) || i < 0 || i > reg_nmatches())
    return __FAIL;
  if (reg_start() >= 0) {
    long pos = reg_pos(i), end = reg_end(i);
    char *s;
    if (pos < 0 || end < 0)
      s = strdup("");
    else if (!(s = malloc(end-pos+1)))
      return __ERROR;
    else {
      strncpy(s, reg_str(i), end-pos);
      s[end-pos] = 0;
    }
    return mkstr(s);
  } else
    return __FAIL;
}

FUNCTION(clib,regpos,argc,argv)
{
  long i;
  if (argc != 1 || !isint(argv[0], &i) || i < 0 || i > reg_nmatches())
    return __FAIL;
  if (reg_start() >= 0)
    return mkint(reg_pos(i));
  else
    return __FAIL;
}

FUNCTION(clib,regend,argc,argv)
{
  long i;
  if (argc != 1 || !isint(argv[0], &i) || i < 0 || i > reg_nmatches())
    return __FAIL;
  if (reg_start() >= 0)
    return mkint(reg_end(i));
  else
    return __FAIL;
}

FUNCTION(clib,regs,argc,argv)
{
  expr x;
  size_t i;
  if (argc != 0) return __FAIL;
  x = mknil;
  i = reg_nmatches();
  while (x && i > 0) {
    if (reg_pos(i) >= 0 && reg_end(i) >= 0)
      x = mkcons(mkint(i), x);
    i--;
  }
  if (x)
    return x;
  else
    return __ERROR;
}

/* additional integer functions: ******************************************/

FUNCTION(clib,pow,argc,argv)
{
  mpz_t m, u;
  unsigned long n;
  if (argc != 2 || !ismpz(argv[0], m) || !isuint(argv[1], &n) ||
      n > 0 && mpz_size(m) > INT_MAX/n)
    return __FAIL;
  if (!mpz_new(u, n*mpz_size(m)))
    return __ERROR;
  mpz_pow_ui(u, m, n);
  if (!mpz_actsize(u))
    return __ERROR;
  return mkmpz(u);
}

FUNCTION(clib,root,argc,argv)
{
  mpz_t m, u;
  unsigned long n;
  if (argc != 2 || !ismpz(argv[0], m) || !isuint(argv[1], &n) || n == 0 ||
      mpz_sgn(m) == -1 && (n&1) == 0)
    return __FAIL;
  if (!mpz_new(u, mpz_size(m)/n+1))
    return __ERROR;
  mpz_root(u, m, n);
  if (!mpz_actsize(u))
    return __ERROR;
  return mkmpz(u);
}

FUNCTION(clib,intsqrt,argc,argv)
{
  mpz_t m, u;
  if (argc != 1 || !ismpz(argv[0], m) || mpz_sgn(m) < 0)
    return __FAIL;
  if (!mpz_new(u, (mpz_size(m)>>2)+1))
    return __ERROR;
  mpz_sqrt(u, m);
  if (!mpz_actsize(u))
    return __ERROR;
  return mkmpz(u);
}

FUNCTION(clib,powmod,argc,argv)
{
  mpz_t k, k2, m, n, u;
  int sgn;
  if (argc != 3 || !ismpz(argv[0], k) || !ismpz(argv[1], m) ||
      !ismpz(argv[2], n) || mpz_sgn(n) < 0 ||
      mpz_sgn(k) == 0)
    return __FAIL;
  if (!mpz_new(u, mpz_size(k)))
    return __ERROR;
  sgn = mpz_sgn(k);
  if (!mpz_copy(k2, k)) { mpz_clear(u); return __ERROR; }
  if (sgn < 0) k2->_mp_size = -k2->_mp_size;
  mpz_powm(u, m, n, k2);
  mpz_clear(k2);
  if (!mpz_actsize(u))
    return __ERROR;
  return mkmpz(u);
}

FUNCTION(clib,invmod,argc,argv)
{
  mpz_t k, m, u;
  if (argc != 2 || !ismpz(argv[0], k) || !ismpz(argv[1], m) ||
      mpz_sgn(k) == 0)
    return __FAIL;
  if (!mpz_new(u, mpz_size(k)))
    return __ERROR;
  if (!mpz_invert(u, m, k)) {
    mpz_clear(u);
    return __FAIL;
  }
  if (!mpz_actsize(u))
    return __ERROR;
  return mkmpz(u);
}

FUNCTION(clib,isprime,argc,argv)
{
  static long rep = 0;
  mpz_t n;
  int ret;
  if (argc != 1 || !ismpz(argv[0], n))
    return __FAIL;
  if (!rep) {
    /* get this value only once when we're first invoked, to avoid the
       overhead in subsequent calls */
    expr x = eval(mksym(sym(ISPRIME_REP)));
    if (x) {
      if (!isint(x, &rep) || rep <= 0) rep = 5;
      dispose(x);
    } else 
      rep = 5;
  }
  ret = mpz_probab_prime_p(n, rep);
  if (ret == 2)
    return mktrue;
  else if (ret == 0)
    return mkfalse;
  else
    return __FAIL;
}

FUNCTION(clib,gcd,argc,argv)
{
  mpz_t m, n, u;
  if (argc != 2 || !ismpz(argv[0], m) || !ismpz(argv[1], n) ||
      mpz_sgn(m) == 0 && mpz_sgn(n) == 0)
    return __FAIL;
  if (!mpz_new(u, long_min(mpz_size(m),mpz_size(n))))
    return __ERROR;
  mpz_gcd(u, m, n);
  if (!mpz_actsize(u))
    return __ERROR;
  return mkmpz(u);
}

FUNCTION(clib,lcm,argc,argv)
{
  mpz_t m, n, u;
  if (argc != 2 || !ismpz(argv[0], m) || !ismpz(argv[1], n) ||
      mpz_sgn(m) == 0 || mpz_sgn(n) == 0 ||
      mpz_size(m)+mpz_size(n) < 0)
    return __FAIL;
  if (!mpz_new(u, mpz_size(m)+mpz_size(n)))
    return __ERROR;
  mpz_lcm(u, m, n);
  if (!mpz_actsize(u))
    return __ERROR;
  return mkmpz(u);
}

FUNCTION(clib,fact,argc,argv)
{
  mpz_t m, n, u;
  unsigned long ret;
  long k;
  if (argc != 2 || !ismpz(argv[0], m) || !ismpz(argv[1], n) ||
      mpz_sgn(m) == 0 || mpz_sgn(n) <= 0 ||
      isint(argv[1], &k) && k == 1)
    return __FAIL;
  if (!mpz_new(u, mpz_size(m)))
    return __ERROR;
  ret = mpz_remove(u, m, n);
  mpz_clear(u);
  return mkuint(ret);
}

FUNCTION(clib,rem,argc,argv)
{
  mpz_t m, n, u;
  long k;
  if (argc != 2 || !ismpz(argv[0], m) || !ismpz(argv[1], n) ||
      mpz_sgn(m) == 0 || mpz_sgn(n) <= 0 ||
      isint(argv[1], &k) && k == 1)
    return __FAIL;
  if (!mpz_new(u, mpz_size(m)))
    return __ERROR;
  mpz_remove(u, m, n);
  if (!mpz_actsize(u))
    return __ERROR;
  return mkmpz(u);
}

FUNCTION(clib,jacobi,argc,argv)
{
  mpz_t m, n;
  int ret;
  if (argc != 2 || !ismpz(argv[0], m) || !ismpz(argv[1], n) ||
      mpz_sgn(n) <= 0)
    return __FAIL;
  ret = mpz_jacobi(m, n);
  return mkint(ret);
}

/* C replacements of common stdlib functions: *****************************/

/* Note: In difference to our usual philosophy to make an operation __FAIL if
   certain implementation limits are exceeded, the following operations return
   an __ERROR in such cases. This is done whenever the operation, if __FAILed,
   would certainly cause some stack or memory overflow later, so we might as
   well abort it right here. */

FUNCTION(clib,append,argc,argv)
{
  if (argc == 2) {
    long n;
    expr x, *ys, hd, tl;
    for (n = 0, x = argv[0]; iscons(x, &hd, &tl); n++)
      if (n >= INT_MAX/sizeof(expr)-1)
	return __ERROR;
      else
	x = tl;
    if (!isnil(x))
      return __FAIL;
    else if (!(ys = malloc((n+1)*sizeof(expr))))
      return __ERROR;
    for (n = 0, x = argv[0]; iscons(x, &hd, &tl); n++) {
      ys[n] = hd;
      x = tl;
    }
    ys[n++] = argv[1];
    return mklistv(n, ys);
  } else
    return __FAIL;
}

FUNCTION(clib,cat,argc,argv)
{
  expr *xs, x, y, hd, tl, hd1, tl1;
  int n;
  if (argc != 1) return __FAIL;
  for (n = 0, x = argv[0]; iscons(x, &hd, &tl); x = tl) {
    for (y = hd; iscons(y, &hd1, &tl1); y = tl1)
      if (n >= INT_MAX/sizeof(expr))
	return __ERROR;
      else
	n++;
    if (!isnil(y)) return __FAIL;
  }
  if (!isnil(x)) return __FAIL;
  if (!(xs = malloc(n*sizeof(expr))))
    return __ERROR;
  for (n = 0, x = argv[0]; iscons(x, &hd, &tl); x = tl)
    for (y = hd; iscons(y, &hd1, &tl1); y = tl1)
      xs[n++] = hd1;
  return mklistv(n, xs);
}

FUNCTION(clib,mklist,argc,argv)
{
  long n;
  if (argc == 2 && isint(argv[1], &n)) {
    expr x = argv[0], y = mknil;
    while (y && n-- > 0)
      y = mkcons(x, y);
    if (!y)
      return __ERROR;
    else
      return y;
  } else
    return __FAIL;
}

FUNCTION(clib,nums,argc,argv)
{
  mpz_t n, m, u;
  double f, g;
  long len, i, j;
  expr *xs;
  unsigned char n_is_mpz, m_is_mpz;

  if (argc != 2 ||
      !((n_is_mpz = ismpz(argv[0], n)) || isfloat(argv[0], &f)) ||
      !((m_is_mpz = ismpz(argv[1], m)) || isfloat(argv[1], &g)))
    return __FAIL;
  if (n_is_mpz) {
    if (!m_is_mpz && !mpz_from_double(m, g))
      return __ERROR;
    if (mpz_cmp(n, m) > 0)
      return mknil;
    if (!mpz_addop2(mpz_sub, u, m, n))
      return __ERROR;
    if (!my_mpz_fits_slong_p(u)) {
      mpz_clear(u);
      return __ERROR;
    }
    len = mpz_get_si(u)+1;
    mpz_clear(u);
    if (len <= 0 || len > INT_MAX/sizeof(expr)) return __ERROR;
    if (!(xs = malloc(len*sizeof(expr))))
      return __ERROR;
    if (!mpz_copy(u, n)) return __ERROR;
    i = 0;
    while (mpz_cmp(u, m) <= 0) {
      mpz_t v;
      if (!(xs[i++] = mkmpz(u))) {
	for (j = 0; j < i-1; j++) dispose(xs[j]);
	free(xs);
	return __ERROR;
      }
      if (!mpz_addop1(mpz_add_ui, v, u, 1U)) {
	for (j = 0; j < i; j++) dispose(xs[j]);
	free(xs);
	mpz_clear(u);
	return __ERROR;
      }
      memcpy(u, v, sizeof(mpz_t));
    }
  } else {
    double ip;
    if (m_is_mpz) g = mpz_get_d(m);
    if (g < f) return mknil;
    modf(g-f, &ip);
    if (ip >= (double)INT_MAX) return __ERROR;
    /* add 1 to be safe (will be corrected later) */
    len = ((int)ip)+2;
    if (len <= 0 || len > INT_MAX/sizeof(expr)) return __ERROR;
    if (!(xs = malloc(len*sizeof(expr))))
      return __ERROR;
    i = 0;
    while (f <= g) {
      if (!(xs[i++] = mkfloat(f))) {
	for (j = 0; j < i-1; j++) dispose(xs[j]);
	free(xs);
	return __ERROR;
      }
      f += 1.0;
    }
    if (i < len) len = i;
  }
  return mklistv(len, xs);
}

FUNCTION(clib,numsby,argc,argv)
{
  mpz_t n, m, k, u;
  double f, g, h;
  long len, i, j;
  expr *xs;
  unsigned char k_is_mpz, n_is_mpz, m_is_mpz;

  if (argc != 3 ||
      !((k_is_mpz = ismpz(argv[0], k)) || isfloat(argv[0], &h)) ||
      !((n_is_mpz = ismpz(argv[1], n)) || isfloat(argv[1], &f)) ||
      !((m_is_mpz = ismpz(argv[2], m)) || isfloat(argv[2], &g)))
    return __FAIL;
  if (n_is_mpz && k_is_mpz) {
    int sgn = mpz_sgn(k);
    if (sgn == 0) return __FAIL;
    if (!m_is_mpz && !mpz_from_double(m, g))
      return __ERROR;
    if (mpz_cmp(n, m)*sgn > 0)
      return mknil;
    if (!mpz_addop2(mpz_sub, u, m, n))
      return __ERROR;
    /* this one shouldn't fail */
    mpz_tdiv_q(u, u, k);
    if (!my_mpz_fits_slong_p(u)) {
      mpz_clear(u);
      return __ERROR;
    }
    len = mpz_get_si(u)+1;
    mpz_clear(u);
    if (len <= 0 || len > INT_MAX/sizeof(expr)) return __ERROR;
    if (!(xs = malloc(len*sizeof(expr))))
      return __ERROR;
    if (!mpz_copy(u, n)) return __ERROR;
    i = 0;
    while (sgn*mpz_cmp(u, m) <= 0) {
      mpz_t v;
      if (!(xs[i++] = mkmpz(u))) {
	for (j = 0; j < i-1; j++) dispose(xs[j]);
	free(xs);
	return __ERROR;
      }
      if (!mpz_addop2(mpz_add, v, u, k)) {
	for (j = 0; j < i; j++) dispose(xs[j]);
	free(xs);
	mpz_clear(u);
	return __ERROR;
      }
      memcpy(u, v, sizeof(mpz_t));
    }
  } else {
    double ip, sgn;
    if (k_is_mpz) h = mpz_get_d(k);
    if (n_is_mpz) f = mpz_get_d(n);
    if (m_is_mpz) g = mpz_get_d(m);
    sgn = (h<0.0)?-1.0:(h>0.0)?1.0:0.0;
    if (sgn == 0.0) return __FAIL;
    if ((f-g)*sgn > 0.0) return mknil;
    modf((g-f)/h, &ip);
    if (ip >= (double)INT_MAX) return __ERROR;
    len = ((int)ip)+2;
    if (len <= 0 || len > INT_MAX/sizeof(expr)) return __ERROR;
    if (!(xs = malloc(len*sizeof(expr))))
      return __ERROR;
    i = 0;
    while (sgn*(f-g) <= 0) {
      if (!(xs[i++] = mkfloat(f))) {
	for (j = 0; j < i-1; j++) dispose(xs[j]);
	free(xs);
	return __ERROR;
      }
      f += h;
    }
    if (i < len) len = i;
  }
  return mklistv(len, xs);
}

FUNCTION(clib,reverse,argc,argv)
{
  if (argc == 1) {
    expr x = argv[0], y = mknil, hd, tl;
    while (y && iscons(x, &hd, &tl)) {
      expr z = mkcons(hd, y);
      y = z; x = tl;
    }
    if (!y)
      return __ERROR;
    else if (isnil(x))
      return y;
    else {
      dispose(y);
      return __FAIL;
    }
  } else
    return __FAIL;
}

FUNCTION(clib,tuplecat,argc,argv)
{
  expr *xs, *ys, x, hd, tl;
  int n, l;
  if (argc != 1) return __FAIL;
  for (n = 0, x = argv[0]; iscons(x, &hd, &tl); x = tl) {
    if (istuple(hd, &l, &ys))
      if (l < 0 || n > INT_MAX/sizeof(expr)-l)
	return __ERROR;
      else
	n += l;
    else
      return __FAIL;
  }
  if (!isnil(x)) return __FAIL;
  if (!(xs = malloc(n*sizeof(expr))))
    return __ERROR;
  for (n = 0, x = argv[0]; iscons(x, &hd, &tl); x = tl)
    if (istuple(hd, &l, &ys) & l > 0) {
      memcpy(xs+n, ys, l*sizeof(expr));
      n += l;
    }
  return mktuplev(n, xs);
}

FUNCTION(clib,chars,argc,argv)
{
  char *s;
  if (argc == 1 && isstr(argv[0], &s)) {
    int l = strlen(s);
    if (l < 0 || l > INT_MAX/sizeof(expr))
      return __ERROR;
    else if (l == 0)
      return mknil;
    else {
      expr *xs = malloc(l*sizeof(expr));
      if (xs) {
	int i = 0, j;
	while(*s) {
	  char *t = malloc(2*sizeof(char));
	  if (t) {
	    t[0] = *(s++);
	    t[1] = 0;
	    if (!(xs[i] = mkstr(t)))
	      goto errexit;
	    else
	      i++;
	  } else
	    goto errexit;
	}
	return mklistv(l, xs);
      errexit:
	for (j = 0; j < i; j++) dispose(xs[j]);
	free(xs);
	return __ERROR;
      } else
	return __ERROR;
    }
  } else
    return __FAIL;
}

FUNCTION(clib,join,argc,argv)
{
  expr x, hd, tl;
  char *s, *t, *delim;
  int n, l, k, init;
  if (argc != 2 || !isstr(argv[0], &delim)) return __FAIL;
  k = strlen(delim);
  for (init = n = 0, x = argv[1]; iscons(x, &hd, &tl); x = tl) {
    if (isstr(hd, &s)) {
      l = strlen(s);
      if (n > 0) {
	if (l < 0 || l >= INT_MAX-k)
	  return __ERROR;
	l += k;
      }
      if (l < 0 || n >= INT_MAX-l)
	return __ERROR;
      else
	n += l;
      init = 1;
    } else
      return __FAIL;
  }
  if (!isnil(x)) return __FAIL;
  if (!(t = malloc(n+1)))
    return __ERROR;
  *t = 0;
  for (init = n = 0, x = argv[1]; iscons(x, &hd, &tl); x = tl)
    if (isstr(hd, &s)) {
      if (init) {
	strcpy(t+n, delim);
	n += k;
      }
      strcpy(t+n, s);
      n += strlen(s);
      init = 1;
    }
  return mkstr(t);
}

/* This is essentially the same algorithm as in string.q, but implemented
   in C for greater efficiency. */

FUNCTION(clib,split,argc,argv)
{
  char *s, *t, *delim;
  expr *xs;
  int n, l, i, j;
  if (argc == 2 && isstr(argv[0], &delim) && isstr(argv[1], &s)) {
    l = strlen(s);
    if (l < 0)
      return __FAIL;
    else if (l == 0)
      return mknil;
    i = j = n = 0;
    while (i < l) {
      if (j < l && !strchr(delim, s[j]))
	j++;
      else {
	n++;
	if (j == l-1) {
	  /* trailing delimiter */
	  n++;
	  break;
	} else
	  i = ++j;
      }
    }
    if (n > INT_MAX/sizeof(expr)) return __ERROR;
    if (!(xs = malloc(n*sizeof(expr)))) return __ERROR;
    i = j = n = 0;
    while (i < l) {
      if (j < l && !strchr(delim, s[j]))
	j++;
      else {
	if (!(t = malloc(j-i+1))) goto errexit;
	strncpy(t, s+i, j-i);
	t[j-i] = 0;
	if (!(xs[n] = mkstr(t))) goto errexit;
	n++;
	if (j == l-1) {
	  /* trailing delimiter */
	  if (!(t = malloc(1))) goto errexit;
	  *t = 0;
	  if (!(xs[n] = mkstr(t))) goto errexit;
	  n++;
	  break;
	} else
	  i = ++j;
      }
    }
    return mklistv(n, xs);
  errexit:
    for (i = 0; i < n; i++) dispose(xs[i]);
    free(xs);
    return __ERROR;
  } else
    return __FAIL;
}

FUNCTION(clib,strcat,argc,argv)
{
  expr x, hd, tl;
  char *s, *t;
  int n, l;
  if (argc != 1) return __FAIL;
  for (n = 0, x = argv[0]; iscons(x, &hd, &tl); x = tl) {
    if (isstr(hd, &s)) {
      l = strlen(s);
      if (l < 0 || n >= INT_MAX-l)
	return __ERROR;
      else
	n += l;
    } else
      return __FAIL;
  }
  if (!isnil(x)) return __FAIL;
  if (!(t = malloc(n+1)))
    return __ERROR;
  for (n = 0, x = argv[0]; iscons(x, &hd, &tl); x = tl)
    if (isstr(hd, &s)) {
      strcpy(t+n, s);
      n += strlen(s);
    }
  return mkstr(t);
}

static expr pred;
static int err;

static int eval_pred(const void *x, const void *y)
{
  expr t = mkapp(mkapp(pred, *((expr*)x)), *((expr*)y)), p;
  if (t && (p = eval(t))) {
    int ret;
    if (istrue(p))
      ret = 1;
    else if (isfalse(p))
      ret = 0;
    else {
      /* values are incomparable */
      err = 1;
      ret = 0;
    }
    return ret;
  } else {
    /* fatal error */
    err = -1;
    return 0;
  }
}

static int cmp_p(const void *x, const void *y)
{
  if (err)
    return 0;
  else if (eval_pred(x, y))
    return -1;
  else if (err)
    return 0;
  else if (eval_pred(y, x))
    return 1;
  else
    return 0;
}

FUNCTION(clib,sort,argc,argv)
{
  if (argc == 2) {
    expr *xs, p = argv[0], x, hd, tl;
    int n;
    for (n = 0, x = argv[1]; iscons(x, &hd, &tl); x = tl) {
      if (n >= INT_MAX/sizeof(expr))
	return __ERROR;
      else
	n++;
    }
    if (!isnil(x)) return __FAIL;
    if (!(xs = malloc(n*sizeof(expr))))
      return __ERROR;
    for (n = 0, x = argv[1]; iscons(x, &hd, &tl); x = tl)
      xs[n++] = hd;
    err = 0;
    pred = p;
    qsort(xs, n, sizeof(expr), cmp_p);
    if (err) {
      free(xs);
      return (err==-1)?__ERROR:__FAIL;
    } else
      return mklistv(n, xs);
  } else
    return __FAIL;
}

#ifdef USE_THREADS

/* do necessary cleanup at fork time */

static void atfork_child(void)
{
  THREAD *thr, *this = thr0+this_thread();
  for (thr = thr0; thr <= thr0+maxused; thr++)
    if (thr->used) {
      pthread_mutex_init(&thr->exit_mutex, NULL);
      pthread_cond_init(&thr->exit_cond, NULL);
      if (thr == this)
	thr->id = pthread_self();
      else {
	thr->active = 0; thr->canceled = 1;
      }
    }
  stamp++;
#ifdef DEBUG
  printf("thread %d (id %d) forked\n", this-thr0, (int)this->id);
#endif
}

#endif

INIT(clib)
{
  thr0->active = 1; thr0->canceled = 0; thr0->used = 1;
  thr0->result = NULL;
  maxused = 0;
#ifdef USE_THREADS
  thr0->id = pthread_self();
  pthread_mutex_init(&thr0->exit_mutex, NULL);
  pthread_cond_init(&thr0->exit_cond, NULL);
  pthread_mutex_init(&format_mutex, NULL);
  thread_atfork(NULL, NULL, atfork_child);
#endif
}

FINI(clib)
{
#ifndef __MINGW32__
  /* clean up zombies */
  int pid, status, serrno;
  serrno = errno;
  while (1) {
    pid = waitpid (-1, &status, WNOHANG);
    if (pid <= 0) break;
  }
  errno = serrno;
#endif
}
